from typing import Dict, List

import altair as alt
import pandas as pd

from edsteva.viz.utils import (
    add_selection_on_legend,
    create_groupby_selection,
    generate_main_chart,
)


def probe_line(
    predictor: pd.DataFrame,
    indexes: List[Dict[str, str]],
    x_axis_title: str,
    y_axis_title: str,
    main_chart_config: Dict[str, float],
):
    """Script to be used by [``plot_probe()``][edsteva.viz.plots.probe.wrapper]

    Parameters
    ----------
    predictor : pd.DataFrame
        $c(t)$ computed in the Probe
    indexes : List[str]
        Variable from which data is grouped
    x_axis_title: str, optional,
        Label name for the x axis.
    y_axis_title: str, optional,
        Label name for the y axis.
    main_chart_config: Dict[str, str], optional
        If not None, configuration used to construct the top main chart.
    """
    base = alt.Chart(predictor)

    index_selection, index_fields = create_groupby_selection(
        indexes=indexes,
        predictor=predictor,
    )
    main_chart = generate_main_chart(
        base=base,
        main_chart_config=main_chart_config,
        index_selection=index_selection,
        index_fields=index_fields,
        x_axis_title=x_axis_title,
        y_axis_title=y_axis_title,
    )
    main_chart = main_chart.mark_line()
    main_chart = add_selection_on_legend(main_chart)

    if index_selection:
        main_chart = main_chart.add_params(index_selection)

    return main_chart



from abc import ABCMeta, abstractmethod
from functools import reduce
from typing import List

import pandas as pd
from loguru import logger

from edsteva import CACHE_DIR
from edsteva.metrics import metrics
from edsteva.probes.base import BaseProbe
from edsteva.probes.utils.filter_df import filter_table_by_date
from edsteva.utils.file_management import delete_object, load_object, save_object


class BaseModel(metaclass=ABCMeta):

    """Base class for Models

    Attributes
    ----------
     _coefs: List[str]
        The list of the Model coefficients
    estimates: pd.DataFrame
        Available with the [``fit()``][edsteva.models.base.BaseModel.fit] method
    _metrics: List[str]
        Available with the [``fit()``][edsteva.models.base.BaseModel.fit] method

        The list of computed metrics if any
    params: List[str]
        Available with the [``fit()``][edsteva.models.base.BaseModel.fit] method

        Ths list of extra keyword parameters used.
    """

    def __init__(
        self,
        algo: str,
        coefs: List[str],
        default_metrics: List[str],
    ):
        self._algo = algo
        self._coefs = coefs
        self._default_metrics = default_metrics
        self._viz_config = {}

    def is_computed_estimates(self) -> None:
        """Raises an error if the Probe has not been fitted properly"""
        if hasattr(self, "estimates"):
            if isinstance(self.estimates, pd.DataFrame):
                if len(self.estimates) == 0:
                    raise Exception(
                        "Estimates are empty, please review the process method or your arguments"
                    )
            else:
                raise Exception(
                    "The fit process must return a Pandas Dataframe and not {}".format(
                        type(self.estimates).__name__
                    )
                )

        else:
            raise Exception(
                "Model has not been fitted, please use the fit method as follow: Model.fit()"
            )

    @abstractmethod
    def fit_process(
        self,
        predictor: pd.DataFrame,
        index: List[str] = None,
        **kwargs,
    ):
        """Fit the Probe in order to obtain estimates"""

    @abstractmethod
    def predict_process(
        self,
        prediction: pd.DataFrame,
        **kwargs,
    ):
        """Compute the predicted Probe"""

    def fit(
        self,
        probe: BaseProbe,
        metric_functions: List[str] = None,
        start_date: str = None,
        end_date: str = None,
        with_cache: bool = True,
        **kwargs,
    ) -> None:
        """Fit the model to the probe instance

        Parameters
        ----------
        probe : BaseProbe
            Target variable to be fitted
        metric_functions : List[str], optional
            Metrics to apply on the fitted Probe. By default it will apply the default metric specified in the model.

            **EXAMPLE**: `[error, error_after_t0]`
        start_date : str, optional
            **EXAMPLE**: `"2019-05-01"`
        end_date : str, optional
            **EXAMPLE**: `"2021-07-01"`

        Examples
        --------
        ```python
        from edsteva.models.step_function import StepFunction

        step_function_model = StepFunction()
        step_function_model.fit(probe)
        step_function_model.estimates.head()
        ```

        | care_site_level          | care_site_id | stay_type    | t_0        | c_0   | error |
        | :----------------------- | :----------- | :----------- | :--------- | :---- | :---- |
        | Unité Fonctionnelle (UF) | 8312056386   | 'Urg_Hospit' | 2019-05-01 | 0.397 | 0.040 |
        | Unité Fonctionnelle (UF) | 8312056386   | 'All'        | 2011-04-01 | 0.583 | 0.028 |
        | Pôle/DMU                 | 8312027648   | 'Urg_Hospit' | 2021-03-01 | 0.677 | 0.022 |
        | Pôle/DMU                 | 8312027648   | 'All'        | 2018-08-01 | 0.764 | 0.014 |
        | Hôpital                  | 8312022130   | 'Urg_Hospit' | 2022-02-01 | 0.652 | 0.027 |
        """
        if isinstance(probe, BaseProbe):
            probe.is_computed_probe()
        else:
            raise TypeError(
                "Unsupported type {} for probe.".format(type(probe).__name__)
            )

        predictor = filter_table_by_date(
            table=probe.predictor,
            table_name="predictor",
            start_date=start_date,
            end_date=end_date,
        )
        index = probe._index

        estimates = self.fit_process(
            predictor=predictor,
            index=index,
            **kwargs,
        )

        metrics_df = self._compute_metrics(
            predictor=predictor,
            estimates=estimates,
            index=index,
            metric_functions=metric_functions,
        )

        if metrics_df is not None:
            self._metrics = list(metrics_df.columns.difference(index))
            self.estimates = estimates.merge(metrics_df, on=index)

        else:
            self.estimates = estimates

        self.is_computed_estimates()
        self.params = kwargs
        if with_cache:
            self.cache_estimates()

    def reset_estimates(
        self,
    ) -> None:
        """Reset the estimates to its initial state"""
        self.estimates = self._cache_estimates.copy()

    def cache_estimates(
        self,
    ) -> None:
        """Cache the predictor"""
        self._cache_estimates = self.estimates.copy()
        logger.info(
            "Cache the estimates, you can reset the estimates to this state with the method reset_estimates"
        )

    def predict(
        self,
        probe: BaseProbe,
    ) -> pd.DataFrame:
        """Computes the predicted probe by using the estimates

        Parameters
        ----------
        probe : BaseProbe
            Target variable to be predicted

        Examples
        --------
        ```python
        from edsteva.models.step_function import StepFunction

        step_function_model.predict(visit).head()
        ```

        | care_site_level          | care_site_id | stay_type    | date       | n_visit | c     | c_fit |
        | :----------------------- | :----------- | :----------- | :--------- | :------ | :---- | :---- |
        | Unité Fonctionnelle (UF) | 8312056386   | 'Urg_Hospit' | 2019-05-01 | 233.0   | 0.841 | 0.758 |
        | Unité Fonctionnelle (UF) | 8312056386   | 'All'        | 2021-04-01 | 393.0   | 0.640 | 0.758 |
        | Pôle/DMU                 | 8312027648   | 'Urg_Hospit' | 2011-03-01 | 204.0   | 0.497 | 0     |
        | Pôle/DMU                 | 8312027648   | 'All'        | 2018-08-01 | 22.0    | 0.784 | 0.874 |
        | Hôpital                  | 8312022130   | 'Urg_Hospit' | 2022-02-01 | 9746.0  | 0.974 | 0.912 |

        """

        predictor = probe.predictor
        index = probe._index

        return self.predict_process(predictor=predictor, index=index)

    def load(self, path=None) -> None:
        """Loads a Model from local

        Parameters
        ----------
        path : str, optional
            **EXAMPLE**: `"my_folder/my_file.html"`

        Examples
        -------
        ```python
        from edsteva.probes import VisitProbe

        probe_path = "my_path/visit.pkl"

        visit = VisitProbe()
        visit.load(path=probe_path)
        ```

        """

        path = path or self._get_path()
        loaded_model = load_object(path)
        self.__dict__ = loaded_model.__dict__.copy()
        self.path = path

    def save(self, path: str = None, name: str = None) -> bool:
        """Saves computed Model instance

        Parameters
        ----------
        path : str, optional
            **EXAMPLE**: `"my_folder/my_file.html"`
        name : str, optional
            **EXAMPLE**: `"fitted_visit"`

        Examples
        -------
        ```python
        from edsteva.probes import VisitProbe

        probe_path = "my_path/visit.pkl"

        visit = VisitProbe()
        visit.compute(data)
        visit.save(path=probe_path)
        ```

        """

        self.is_computed_estimates()

        if name:
            self.name = name
        if not path:
            path = self._get_path()

        self.path = path
        save_object(self, path)

    def delete(self, path: str = None) -> bool:
        """Delete the saved Model instance

        Parameters
        ----------
        path : str, optional
            **EXAMPLE**: `"my_folder/my_file.html"`
        """
        if not path:
            path = self.path

        delete_object(self, path)

    def _get_path(self):
        base_path = CACHE_DIR / "edsteva" / "models"
        if hasattr(self, "name"):
            filename = f"{self.name.lower()}.pickle"
        else:
            filename = f"{type(self).__name__.lower()}.pickle"
        return base_path / filename

    def _compute_metrics(
        self,
        predictor: pd.DataFrame,
        estimates: pd.DataFrame,
        index: List[str],
        metric_functions: List[str] = None,
    ):
        if metric_functions is None:
            if hasattr(self, "_default_metrics") and self._default_metrics:
                metric_functions = self._default_metrics
            else:
                return None
        if isinstance(metric_functions, str):
            metric_functions = [metric_functions]
        metrics_df = []
        for metric_function in metric_functions:
            metrics_df.append(
                metrics.get(metric_function)(
                    predictor=predictor, estimates=estimates, index=index
                )
            )
        return reduce(lambda left, right: left.merge(right, on=index), metrics_df)

    def is_predictable_probe(
        self,
        predictor: pd.DataFrame,
        index: List[str],
    ) -> pd.DataFrame:
        """Raises an error if the model has not been fitted on the input predictor.

        Parameters
        ----------
        predictor : pd.DataFrame
            Target DataFrame to be predicted
        index : List[str]
            List of the columns given by Probe._index

        Returns
        -------
        pd.DataFrame
            Predictor along with the fitted estimates

        Raises
        ------
        Exception
            Some indexes have no associated estimates, the model must be fitted on an adequate probe
        """
        prediction = predictor.merge(
            self.estimates, on=index, how="left", validate="many_to_one", indicator=True
        )
        if (prediction["_merge"] == "both").all():
            return prediction.drop(columns="_merge")

        raise Exception(
            "Some indexes have no associated estimates, the model must be fitted on an adequate probe"
        )



from datetime import datetime
from typing import Dict, List, Union

import pandas as pd
from loguru import logger

from edsteva.probes.utils.filter_df import convert_uf_to_pole
from edsteva.probes.utils.prepare_df import (
    prepare_care_site,
    prepare_condition_occurrence,
    prepare_visit_detail,
    prepare_visit_occurrence,
)
from edsteva.probes.utils.utils import (
    CARE_SITE_LEVEL_NAMES,
    concatenate_predictor_by_level,
    hospital_only,
    impute_missing_dates,
)
from edsteva.utils.checks import check_condition_source_systems, check_tables
from edsteva.utils.framework import is_koalas, to
from edsteva.utils.typing import Data, DataFrame


def compute_completeness_predictor_per_visit(
    self,
    data: Data,
    care_site_relationship: pd.DataFrame,
    start_date: datetime,
    end_date: datetime,
    care_site_levels: List[str],
    stay_types: Union[str, Dict[str, str]],
    care_site_ids: List[int],
    extra_data: Data,
    care_site_short_names: List[str],
    care_site_specialties: List[str],
    care_sites_sets: Union[str, Dict[str, str]],
    specialties_sets: Union[str, Dict[str, str]],
    diag_types: Union[str, Dict[str, str]],
    condition_types: Union[str, Dict[str, str]],
    source_systems: List[str],
    length_of_stays: List[float],
    **kwargs
):
    r"""Script to be used by [``compute()``][edsteva.probes.base.BaseProbe.compute]

    The ``per_visit`` algorithm computes $c_(t)$ the availability of claim data linked to patients' administrative stays:

    $$
    c(t) = \frac{n_{with\,condition}(t)}{n_{visit}(t)}
    $$

    Where $n_{visit}(t)$ is the number of administrative stays, $n_{with\,condition}$ the number of stays having at least one claim code (e.g. ICD-10) recorded and $t$ is the month.
    """

    self._metrics = ["c", "n_visit", "n_visit_with_condition"]
    check_tables(data=data, required_tables=["condition_occurrence"])
    check_condition_source_systems(source_systems=source_systems)
    if "AREM" in source_systems and not hospital_only(
        care_site_levels=care_site_levels
    ):  # pragma: no cover
        logger.info("AREM claim data are only available at hospital level")

    visit_occurrence = prepare_visit_occurrence(
        data=data,
        start_date=start_date,
        end_date=end_date,
        stay_types=stay_types,
        length_of_stays=length_of_stays,
    )

    condition_occurrence = prepare_condition_occurrence(
        data=data,
        extra_data=extra_data,
        visit_occurrence=visit_occurrence,
        source_systems=source_systems,
        diag_types=diag_types,
        condition_types=condition_types,
    ).drop(columns=["condition_occurrence_id", "date"])

    care_site = prepare_care_site(
        data=data,
        care_site_ids=care_site_ids,
        care_site_short_names=care_site_short_names,
        care_site_relationship=care_site_relationship,
        care_site_specialties=care_site_specialties,
        care_sites_sets=care_sites_sets,
        specialties_sets=specialties_sets,
    )

    hospital_visit = get_hospital_visit(
        condition_occurrence,
        visit_occurrence,
        care_site,
    )
    hospital_name = CARE_SITE_LEVEL_NAMES["Hospital"]
    condition_predictor_by_level = {hospital_name: hospital_visit}

    # UF selection
    if not hospital_only(care_site_levels=care_site_levels):
        visit_detail = prepare_visit_detail(data, start_date, end_date)

        uf_visit = get_uf_visit(
            condition_occurrence=condition_occurrence,
            visit_occurrence=visit_occurrence,
            visit_detail=visit_detail,
            care_site=care_site,
        )
        uf_name = CARE_SITE_LEVEL_NAMES["UF"]
        condition_predictor_by_level[uf_name] = uf_visit

        pole_visit = get_pole_visit(
            uf_visit=uf_visit,
            care_site=care_site,
            care_site_relationship=care_site_relationship,
        )
        pole_name = CARE_SITE_LEVEL_NAMES["Pole"]
        condition_predictor_by_level[pole_name] = pole_visit

    condition_predictor = concatenate_predictor_by_level(
        predictor_by_level=condition_predictor_by_level,
        care_site_levels=care_site_levels,
    )

    return compute_completeness(self, condition_predictor)


def compute_completeness(
    self,
    condition_predictor: DataFrame,
):
    # Visit with diagnosis
    partition_cols = [*self._index.copy(), "date"]
    n_visit_with_condition = (
        condition_predictor.groupby(
            partition_cols,
            as_index=False,
            dropna=False,
        )
        .agg({"has_condition": "count"})
        .rename(columns={"has_condition": "n_visit_with_condition"})
    )
    n_visit_with_condition = to("pandas", n_visit_with_condition)
    n_visit_with_condition = n_visit_with_condition[
        n_visit_with_condition.n_visit_with_condition > 0
    ]
    n_visit_with_condition = impute_missing_dates(
        start_date=self.start_date,
        end_date=self.end_date,
        predictor=n_visit_with_condition,
        partition_cols=partition_cols,
    )

    # Visit total
    condition_columns = ["diag_type", "condition_type", "source_system"]
    partition_cols = list(set(partition_cols) - set(condition_columns))
    n_visit = (
        condition_predictor.groupby(
            partition_cols,
            as_index=False,
            dropna=False,
        )
        .agg({"visit_id": "nunique"})
        .rename(columns={"visit_id": "n_visit"})
    )
    n_visit = to("pandas", n_visit)
    n_visit = impute_missing_dates(
        start_date=self.start_date,
        end_date=self.end_date,
        predictor=n_visit,
        partition_cols=partition_cols,
    )

    condition_predictor = n_visit_with_condition.merge(
        n_visit,
        on=partition_cols,
    )

    condition_predictor["c"] = condition_predictor["n_visit"].where(
        condition_predictor["n_visit"] == 0,
        condition_predictor["n_visit_with_condition"] / condition_predictor["n_visit"],
    )

    return condition_predictor


def get_hospital_visit(
    condition_occurrence: DataFrame,
    visit_occurrence: DataFrame,
    care_site: DataFrame,
):
    condition_hospital = condition_occurrence.drop_duplicates(
        ["visit_occurrence_id", "diag_type", "condition_type", "source_system"]
    )
    condition_hospital["has_condition"] = True
    hospital_visit = visit_occurrence.merge(
        condition_hospital,
        on="visit_occurrence_id",
        how="left",
    )
    hospital_visit = hospital_visit.rename(columns={"visit_occurrence_id": "visit_id"})
    hospital_visit = hospital_visit.merge(care_site, on="care_site_id")

    if is_koalas(hospital_visit):
        hospital_visit = hospital_visit.spark.cache()

    return hospital_visit


def get_uf_visit(
    condition_occurrence: DataFrame,
    visit_occurrence: DataFrame,
    visit_detail: DataFrame,
    care_site: DataFrame,
):  # pragma: no cover
    visit_detail = visit_detail[visit_detail.visit_detail_type == "RUM"]
    condition_uf = (
        condition_occurrence[
            [
                "visit_detail_id",
                "diag_type",
                "condition_type",
                "source_system",
            ]
        ]
        .drop_duplicates()
        .rename(columns={"visit_detail_id": "visit_id"})
    )
    condition_uf["has_condition"] = True

    visit_detail = visit_detail.merge(
        visit_occurrence[
            visit_occurrence.columns.intersection(
                set(["visit_occurrence_id", "length_of_stay", "stay_type"])
            )
        ],
        on="visit_occurrence_id",
    )
    uf_visit = visit_detail.merge(
        condition_uf,
        on="visit_id",
        how="left",
    ).drop(columns=["visit_occurrence_id"])

    uf_visit = uf_visit.merge(care_site, on="care_site_id")

    uf_name = CARE_SITE_LEVEL_NAMES["UF"]
    uf_visit = uf_visit[uf_visit["care_site_level"] == uf_name]

    if is_koalas(uf_visit):
        uf_visit = uf_visit.spark.cache()

    return uf_visit


def get_pole_visit(
    uf_visit: DataFrame,
    care_site: DataFrame,
    care_site_relationship: DataFrame,
):  # pragma: no cover
    care_site_cols = list(
        set(
            [
                "care_site_short_name",
                "care_site_level",
                "care_site_specialty",
                "specialties_set",
                "care_sites_set",
            ]
        ).intersection(uf_visit.columns)
    )
    pole_visit = convert_uf_to_pole(
        table=uf_visit.drop(columns=care_site_cols),
        table_name="uf_visit",
        care_site_relationship=care_site_relationship,
    )

    pole_visit = pole_visit.merge(care_site, on="care_site_id")
    pole_name = CARE_SITE_LEVEL_NAMES["Pole"]
    pole_visit = pole_visit[pole_visit["care_site_level"] == pole_name]
    if is_koalas(pole_visit):
        pole_visit = pole_visit.spark.cache()

    return pole_visit



from typing import List

from edsteva.utils.typing import Data, DataFrame


class MissingColumnError(Exception):
    """Exception raised when a concept is missing"""

    def __init__(
        self,
        required_columns: List,
        df_name: str = "",
    ):
        to_display_per_column = [f"- {column}" for column in required_columns]
        str_to_display = "\n".join(to_display_per_column)

        if df_name:
            df_name = f" {df_name} "
        message = (
            f"The{df_name}DataFrame is missing some columns, "
            "namely:\n"
            f"{str_to_display}"
        )

        super().__init__(message)


class MissingTableError(Exception):
    """Exception raised when a table is missing in the Data"""

    def __init__(
        self,
        required_tables: List,
    ):
        to_display_per_concept = [f"- {concept}" for concept in required_tables]
        str_to_display = "\n".join(to_display_per_concept)

        message = f"Data is missing some tables, namely:\n {str_to_display}"

        super().__init__(message)


def check_columns(df: DataFrame, required_columns: List[str], df_name: str = ""):
    present_columns = set(df.columns)
    missing_columns = set(required_columns) - present_columns
    if missing_columns:
        raise MissingColumnError(missing_columns, df_name=df_name)


def check_tables(data: Data, required_tables: List[str]):
    if isinstance(data, Data.__args__):
        present_tables = set(data.available_tables)
        missing_tables = set(required_tables) - present_tables
        if missing_tables:
            raise MissingTableError(missing_tables)
    else:
        raise AttributeError(
            "data should be a Data type please refer to this [page](https://aphp.github.io/edsteva/latest/components/loading_data/)"
        )


def check_condition_source_systems(
    source_systems: List[str], valid_source_systems: List[str] = ["AREM", "ORBIS"]
):
    if source_systems and isinstance(source_systems, list):
        valid = False
        for valid_source_system in valid_source_systems:
            if valid_source_system in source_systems:
                valid = True

        if not valid:
            raise AttributeError(
                "Source systems only accept {}".format(valid_source_systems)
            )
    else:
        raise AttributeError(
            "Source systems must be a non empty list and not {}".format(
                type(source_systems).__name__
            )
        )



from datetime import datetime
from typing import Dict, List, Tuple, Union

import pandas as pd

from edsteva.probes.base import BaseProbe
from edsteva.probes.biology.completeness_predictors import completeness_predictors
from edsteva.probes.biology.viz_configs import viz_configs
from edsteva.utils.typing import Data


class BiologyProbe(BaseProbe):
    r"""
    The ``BiologyProbe`` computes $c_(t)$ the availability of laboratory data related to biological measurements for each biological code and each care site according to time:

    $$
    c(t) = \frac{n_{biology}(t)}{n_{99}}
    $$

    Where $n_{biology}(t)$ is the number of biological measurements, $n_{99}$ is the $99^{th}$ percentile and $t$ is the month.

    Parameters
    ----------
    completeness_predictor: str
        Algorithm used to compute the completeness predictor
        **EXAMPLE**: ``"per_visit_default"``
    standard_terminologies: List[str]
        List of standards terminologies to consider
        **EXAMPLE**: ``["LOINC", "ANABIO"]``

    Attributes
    ----------
    _completeness_predictor: str
        Algorithm used to compute the completeness predictor
        **VALUE**: ``"per_visit_default"``
    _standard_terminologies: List[str]
        List of standards terminologies to consider
        **VALUE**: ``["LOINC", "ANABIO"]``
    _index: List[str]
        Variable from which data is grouped
        **VALUE**: ``["care_site_level", "concepts_set", "stay_type", "length_of_stay", "care_site_id", "care_site_specialty", "specialties_set", "<std_terminology>_concept_code"]``
    _viz_config: List[str]
        Dictionary of configuration for visualization purpose.
        **VALUE**: ``{}``
    """

    def __init__(
        self,
        completeness_predictor: str = "per_measurement_default",
        standard_terminologies: List[str] = ["ANABIO", "LOINC"],
    ):
        self._standard_terminologies = standard_terminologies
        self._index = [
            "care_site_level",
            "concepts_set",
            "stay_type",
            "length_of_stay",
            "care_site_id",
            "care_site_specialty",
            "care_sites_set",
            "specialties_set",
        ] + [
            "{}_concept_code".format(terminology)
            for terminology in standard_terminologies
        ]
        super().__init__(
            completeness_predictor=completeness_predictor,
            index=self._index,
        )

    def compute_process(
        self,
        data: Data,
        care_site_relationship: pd.DataFrame,
        start_date: datetime,
        end_date: datetime,
        care_site_levels: List[str],
        stay_types: Union[str, Dict[str, str]],
        care_site_ids: List[int],
        care_site_short_names: List[str] = None,
        care_site_specialties: List[str] = None,
        concept_codes: List[str] = None,
        care_sites_sets: Union[str, Dict[str, str]] = None,
        specialties_sets: Union[str, Dict[str, str]] = None,
        concepts_sets: Union[str, Dict[str, str]] = {
            "Leucocytes": "A0174|K3232|H6740|E4358|C9784|C8824|E6953",
            "Plaquettes": "E4812|C0326|A1636|A0230|H6751|A1598|G7728|G7727|G7833|A2538|A2539|J4463",
            "Créatinine": "E3180|G1974|J1002|A7813|A0094|G1975|J1172|G7834|F9409|F9410|C0697|H4038|F2621",
            "Potassium": "A1656|C8757|C8758|A2380|E2073|L5014|F2618|E2337|J1178|A3819|J1181",
            "Sodium": "A1772|C8759|C8760|A0262|J1177|F8162|L5013|F2617|K9086|J1180",
            "Chlorure": "B5597|F2359|A0079|J1179|F2619|J1182|F2358|A0079|J1179|F2619|J1182",
            "Glucose": "A1245|E7961|C8796|H7753|A8029|H7749|A0141|H7323|J7401|F2622|B9553|C7236|E7312|G9557|A7338|H7324|C0565|E9889|A8424|F6235|F5659|F2406",
            "Bicarbonate": "A0422|H9622|C6408|F4161",
        },
        length_of_stays: List[float] = [1],
        source_terminologies: Dict[str, str] = {
            "ANALYSES_LABORATOIRE": r"Analyses Laboratoire",
            "GLIMS_ANABIO": r"GLIMS.{0,20}Anabio",
            "GLIMS_LOINC": r"GLIMS.{0,20}LOINC",
            "ANABIO_ITM": r"ITM - ANABIO",
            "LOINC_ITM": r"ITM - LOINC",
        },
        mapping: List[Tuple[str, str, str]] = [
            ("ANALYSES_LABORATOIRE", "GLIMS_ANABIO", "Maps to"),
            ("ANALYSES_LABORATOIRE", "GLIMS_LOINC", "Maps to"),
            ("GLIMS_ANABIO", "ANABIO_ITM", "Mapped from"),
            ("ANABIO_ITM", "LOINC_ITM", "Maps to"),
        ],
        **kwargs,
    ):
        """Script to be used by [``compute()``][edsteva.probes.base.BaseProbe.compute]

        Parameters
        ----------
        data : Data
            Instantiated [``HiveData``][edsteva.io.hive.HiveData], [``PostgresData``][edsteva.io.postgres.PostgresData] or [``LocalData``][edsteva.io.files.LocalData]
        care_site_relationship : pd.DataFrame
            DataFrame computed in the [``compute()``][edsteva.probes.base.BaseProbe.compute] that gives the hierarchy of the care site structure.
        start_date : datetime, optional
            **EXAMPLE**: `"2019-05-01"`
        end_date : datetime, optional
            **EXAMPLE**: `"2021-07-01"`
        care_site_levels : List[str], optional
            **EXAMPLE**: `["Hospital", "Pole", "UF"]`
        stay_types : Union[str, Dict[str, str]], optional
            **EXAMPLE**: `{"All": ".*"}` or `{"All": ".*", "Urg_and_consult": "urgences|consultation"}` or `"hospitalisés`
        care_site_ids : List[int], optional
            **EXAMPLE**: `[8312056386, 8312027648]`
        care_site_short_names : List[str], optional
            **EXAMPLE**: `["HOSPITAL 1", "HOSPITAL 2"]`
        care_site_specialties : List[str], optional
            **EXAMPLE**: `["CARDIOLOGIE", "CHIRURGIE"]`
        concept_codes : List[str], optional
            **EXAMPLE**: ['E3180', 'G1974', 'J1002', 'A7813', 'A0094', 'G1975', 'J1172', 'G7834', 'F9409', 'F9410', 'C0697', 'H4038']`
        care_sites_sets : Union[str, Dict[str, str]], optional
            **EXAMPLE**: `{"All AP-HP": ".*"}` or `{"All AP-HP": ".*", "Pediatrics": r"debre|trousseau|necker"}`
        specialties_sets : Union[str, Dict[str, str]], optional
            **EXAMPLE**: `{"All": ".*"}` or `{"All": ".*", "ICU": r"REA\s|USI\s|SC\s"}`
        concepts_sets : Union[str, Dict[str, str]], optional
            **EXAMPLE**: `{"Créatinine": "E3180|G1974|J1002|A7813|A0094|G1975|J1172|G7834|F9409|F9410|C0697|H4038|F2621", "Leucocytes": r"A0174|K3232|H6740|E4358|C9784|C8824|E6953"}`
        length_of_stays : List[float], optional
            **EXAMPLE**: `[1, 30]`
        source_terminologies : Dict[str, str], optional
            Dictionary of regex used to detect terminology in the column `vocabulary_id`.
            **EXAMPLE**: `{"GLIMS_LOINC": r"GLIMS.{0,20}LOINC"}`
        mapping : List[Tuple[str, str, str]], optional
            List of values to filter in the column `relationship_id` in order to map between 2 terminologies.
            **EXAMPLE**: `[("ANALYSES_LABORATOIRE", "GLIMS_ANABIO", "Maps to")]`
        """
        if specialties_sets is None and "specialties_set" in self._index:
            self._index.remove("specialties_set")
        if length_of_stays is None and "length_of_stay" in self._index:
            self._index.remove("length_of_stay")
        if care_sites_sets is None and "care_sites_set" in self._index:
            self._index.remove("care_sites_set")
        if concepts_sets is None and "concepts_set" in self._index:
            self._index.remove("concepts_set")
        else:
            for terminology in self._standard_terminologies:
                if "{}_concept_code".format(terminology) in self._index:
                    self._index.remove("{}_concept_code".format(terminology))

        return completeness_predictors.get(self._completeness_predictor)(
            self,
            data=data,
            care_site_relationship=care_site_relationship,
            start_date=start_date,
            end_date=end_date,
            care_site_levels=care_site_levels,
            stay_types=stay_types,
            care_site_ids=care_site_ids,
            care_site_short_names=care_site_short_names,
            care_site_specialties=care_site_specialties,
            concept_codes=concept_codes,
            care_sites_sets=care_sites_sets,
            specialties_sets=specialties_sets,
            concepts_sets=concepts_sets,
            length_of_stays=length_of_stays,
            source_terminologies=source_terminologies,
            mapping=mapping,
            **kwargs,
        )

    def get_viz_config(self, viz_type: str, **kwargs):
        if viz_type in viz_configs.keys():
            _viz_config = self._viz_config.get(viz_type)
            if _viz_config is None:
                _viz_config = self._completeness_predictor
        else:
            raise ValueError(f"edsteva has no {viz_type} registry !")
        return viz_configs[viz_type].get(_viz_config)(self, **kwargs)

    def available_completeness_predictors(self):
        return list(completeness_predictors.get_all().keys())



from pathlib import Path
from typing import Dict, Iterable, List, Optional, Union

import pandas
from databricks import koalas
from loguru import logger
from pyspark.sql import DataFrame as SparkDataFrame
from pyspark.sql import SparkSession
from pyspark.sql.types import LongType, StructField, StructType

from edsteva import koalas_options

from . import settings
from .i2b2_mapping import get_i2b2_table

DataFrame = Union[koalas.DataFrame, pandas.DataFrame]


class HiveData:  # pragma: no cover
    """Spark interface for OMOP data stored in a Hive database.

    This class provides a simple access to data stored in Hive.
    Data is returned as koalas dataframes that match the tables
    stored in Hive.

    Parameters
    ----------
    database_name : str
        The name of you database in Hive. Ex: "cse_82727572"
    database_type : str
        The type of your database. Must be "OMOP" or "I2B2"
    spark_session : pyspark.sql.SparkSession, optional
        If None, a SparkSession will be retrieved or  created via `SparkSession.builder.enableHiveSupport().getOrCreate()`
    person_ids : Optional[Iterable[int]], default: None
        An iterable of `person_id` that is used to define a subset of the database.
    tables_to_load : Optional[Dict[str, Optional[List[str]]]], default: None
        By default (i.e. if ``tables_to_load is None``), loaded tables and columns loaded in each table are those listed in
        :py:data:`~edsteva.io.settings.tables_to_load`.
        A dictionnary can be provided to complement those default settings. Keys should be table names to load,
        and values should be:
        - ``None`` to load all columns
        - A list of columns to load (or to add to the default loaded columns if the table is already loaded by default)


    Attributes
    ----------
    person : koalas dataframe
        Hive data for table `person` as a koalas dataframe.
        Other OMOP tables can also be accessed as attributes
    available_tables : list of str
        names of OMOP tables that can be accessed as attributes with this
        HiveData object.


    Examples
    --------

    ```python
    data = HiveData(database_name="edsomop_prod_a")
    data.available_tables
    # Out: ["person", "care_site", "condition_occurrence", ... ]

    person = data.person
    type(person)
    # Out: databricks.koalas.frame.DataFrame

    person["person_id"].count()
    # Out: 12670874
    ```

    This class can be used to create a subset of data for a given
    list of `person_id`. This is useful because the smaller dataset
    can then be used to prototype more rapidly.

    ```python
    my_person_ids = [9226726, 2092082, 5097816]
    data = HiveData(
        spark_session=spark, database_name="edsomop_prod_a", person_ids=my_person_ids
    )
    data.person["person_id"].count()
    # Out: 1000

    tables_to_save = ["person", "visit_occurrence"]
    data.persist_tables_to_folder("./cohort_sample_1000", table_names=tables_to_save)
    # Out: writing /export/home/USER/cohort_sample_1000/person.parquet
    # Out: writing /export/home/USER/cohort_sample_1000/visit_occurrence.parquet
    # Out: ...
    ```

    """

    def __init__(
        self,
        database_name: str,
        database_type: str = "OMOP",
        spark_session: Optional[SparkSession] = None,
        person_ids: Optional[Iterable[int]] = None,
        tables_to_load: Optional[
            Union[Dict[str, Optional[List[str]]], List[str]]
        ] = None,
    ):
        if spark_session is not None:
            self.spark_session = spark_session
        else:
            logger.warning(
                """
                To improve performances when using Spark and Koalas, please call `edsteva.improve_performances()`
                This function optimally configures Spark. Use it as:
                `spark, sc, sql = edsteva.improve_performances()`
                The functions respectively returns a SparkSession, a SparkContext and an sql method
                """
            )
            self.spark_session = SparkSession.builder.enableHiveSupport().getOrCreate()

        koalas_options()

        self.database_name = database_name
        self.database_type = database_type

        if self.database_type == "I2B2":
            self.database_source = "cse" if "cse" in self.database_name else "edsprod"
            self.omop_to_i2b2 = settings.i2b2_tables[self.database_source]
            self.i2b2_to_omop = {}
            for omop_col, i2b2_col in self.omop_to_i2b2.items():
                if i2b2_col in self.i2b2_to_omop.keys():
                    self.i2b2_to_omop[i2b2_col].append(omop_col)
                else:
                    self.i2b2_to_omop[i2b2_col] = [omop_col]

        self.person_ids = self._prepare_person_ids(person_ids)

        tmp_tables_to_load = settings.tables_to_load

        if isinstance(tables_to_load, dict):
            for table_name, columns in tables_to_load.items():
                if columns is None:
                    tmp_tables_to_load[table_name] = None
                else:
                    tmp_tables_to_load[table_name] = list(
                        set(tmp_tables_to_load.get(table_name, []) + columns)
                    )
        elif isinstance(tables_to_load, list):
            for table_name in tables_to_load:
                tmp_tables_to_load[table_name] = None

        self.tables_to_load = tmp_tables_to_load
        self.available_tables = self.list_available_tables()

    def list_available_tables(self) -> List[str]:
        tables_df = self.spark_session.sql(
            f"SHOW TABLES IN {self.database_name}"
        ).toPandas()
        available_tables = []
        for table_name in tables_df["tableName"].drop_duplicates().to_list():
            if (
                self.database_type == "OMOP"
                and table_name in self.tables_to_load.keys()
            ):
                available_tables.append(table_name)
            elif (
                self.database_type == "I2B2" and table_name in self.i2b2_to_omop.keys()
            ):
                for omop_table in self.i2b2_to_omop[table_name]:
                    if omop_table in self.tables_to_load.keys():
                        available_tables.append(omop_table)
            available_tables = list(set(available_tables))
        return available_tables

    def rename_table(self, old_table_name: str, new_table_name: str) -> None:
        if old_table_name in self.available_tables:
            setattr(self, new_table_name, getattr(self, old_table_name))
            self.available_tables.remove(old_table_name)
            self.available_tables.append(new_table_name)
            logger.info("Table {} has been renamed {}", old_table_name, new_table_name)
        else:
            logger.info("Table {} is not available", old_table_name)

    def add_table(self, table_name: str, columns: List[str]) -> None:
        tables_df = self.spark_session.sql(
            f"SHOW TABLES IN {self.database_name}"
        ).toPandas()
        if table_name in tables_df["tableName"].drop_duplicates().to_list():
            self.tables_to_load[table_name] = list(
                set(self.tables_to_load.get(table_name, []) + columns)
            )
            self.available_tables = self.list_available_tables()
            logger.info("Table {} has been added", table_name)

        else:
            raise AttributeError(
                f"Table '{table_name}' is not in the database '{self.database_name}'"
            )

    def delete_table(self, table_name: str) -> None:
        self.tables_to_load.pop(table_name, None)
        self.available_tables = self.list_available_tables()
        logger.info("Table {} has been deleted", table_name)

    def _prepare_person_ids(self, list_of_person_ids) -> Optional[SparkDataFrame]:
        if list_of_person_ids is None:
            return None
        if hasattr(list_of_person_ids, "to_list"):
            # Useful when list_of_person_ids are Koalas (or Pandas) Series
            unique_ids = set(list_of_person_ids.to_list())
        else:
            unique_ids = set(list_of_person_ids)

        logger.info("Number of unique patients: {}", len(unique_ids))
        schema = StructType([StructField("person_id", LongType(), True)])

        return self.spark_session.createDataFrame(
            [(int(p),) for p in unique_ids], schema=schema
        )

    def _read_table(self, table_name, person_ids=None) -> DataFrame:
        assert table_name in self.available_tables
        if person_ids is None and self.person_ids is not None:
            person_ids = self.person_ids

        if self.database_type == "OMOP":
            df = self.spark_session.sql(
                f"select * from {self.database_name}.{table_name}"
            )

        elif self.database_type == "I2B2":
            df = get_i2b2_table(
                spark_session=self.spark_session,
                db_name=self.database_name,
                db_source=self.database_source,
                table=table_name,
            )

        desired_columns = self.tables_to_load[table_name]
        selected_columns = (
            df.columns
            if desired_columns is None
            else [col for col in df.columns if col in desired_columns]
        )
        df = df.select(*selected_columns)

        if "person_id" in df.columns and person_ids is not None:
            df = df.join(person_ids, on="person_id", how="inner")

        return df.to_koalas()

    def persist_tables_to_folder(
        self,
        folder: str,
        person_ids: Optional[Iterable[int]] = None,
        tables: List[str] = None,
    ) -> None:
        """Save OMOP tables as parquet files in a given folder.

        Parameters
        ----------
        folder : str
            path to folder where the tables will be written.
        person_ids : iterable
            person_ids to keep in the subcohort
        tables : list of str, default None
            list of table names to save. Default value is
            :py:data:`~edsteva.io.settings.default_tables_to_save`

        """
        if tables is None:
            tables = settings.default_tables_to_save

        unknown_tables = [
            table for table in tables if table not in self.available_tables
        ]
        if unknown_tables:
            raise ValueError(
                f"The following tables are not available : {unknown_tables}"
            )

        folder = Path.resolve(folder)
        assert Path.exists(folder) and Path.is_dir(
            folder
        ), f"Folder {folder} not found."

        if person_ids is not None:
            person_ids = self._prepare_person_ids(person_ids)

        for table in tables:
            filepath = folder / f"{table}.parquet"
            df = self._read_table(table, person_ids=person_ids)
            self._write_df_to_parquet(df, filepath)

    def _write_df_to_parquet(
        self,
        df: DataFrame,
        filepath: str,
    ) -> None:
        assert Path.is_absolute(filepath)
        logger.info(f"writing {filepath}")
        spark_filepath = "file://" + filepath
        df.to_parquet(spark_filepath, mode="overwrite")

    def __getattr__(self, table_name: str) -> DataFrame:
        if table_name in self.available_tables:
            # the first time it is called, we actually set the attribute
            table = self._read_table(table_name)
            setattr(self, table_name, table)
            return getattr(self, table_name)
        raise AttributeError(f"Table '{table_name}' unknown")

    def __dir__(self) -> List[str]:
        return list(set(list(super().__dir__()) + self.available_tables))



from datetime import datetime
from typing import Dict, List, Union

import pandas as pd

from edsteva.probes.utils.filter_df import convert_uf_to_pole
from edsteva.probes.utils.prepare_df import (
    prepare_care_site,
    prepare_visit_detail,
    prepare_visit_occurrence,
)
from edsteva.probes.utils.utils import (
    CARE_SITE_LEVEL_NAMES,
    VISIT_DETAIL_TYPE,
    concatenate_predictor_by_level,
    hospital_only,
    impute_missing_dates,
)
from edsteva.utils.framework import is_koalas, to
from edsteva.utils.typing import Data, DataFrame


def compute_completeness_predictor_per_visit(
    self,
    data: Data,
    care_site_relationship: pd.DataFrame,
    start_date: datetime,
    end_date: datetime,
    care_site_levels: List[str],
    stay_types: Union[str, Dict[str, str]],
    care_site_ids: List[int],
    care_site_short_names: List[str],
    care_site_specialties: List[str],
    care_sites_sets: Union[str, Dict[str, str]],
    specialties_sets: Union[str, Dict[str, str]],
    length_of_stays: List[float],
    **kwargs
):
    r"""Script to be used by [``compute()``][edsteva.probes.base.BaseProbe.compute]

    The ``per_visit`` algorithm computes $c_(t)$ the availability of administrative data related to visits for each care site according to time:

    $$
    c(t) = \frac{n_{visit}(t)}{n_{max}}
    $$

    Where $n_{visit}(t)$ is the number of administrative stays, $t$ is the month and $n_{max} = \max_{t}(n_{visit}(t))$.
    """
    self._metrics = ["c", "n_visit"]
    visit_occurrence = prepare_visit_occurrence(
        data=data,
        start_date=start_date,
        end_date=end_date,
        stay_types=stay_types,
        length_of_stays=length_of_stays,
    )

    care_site = prepare_care_site(
        data=data,
        care_site_ids=care_site_ids,
        care_site_short_names=care_site_short_names,
        care_site_specialties=care_site_specialties,
        care_site_relationship=care_site_relationship,
        specialties_sets=specialties_sets,
        care_sites_sets=care_sites_sets,
    )

    hospital_visit = get_hospital_visit(
        visit_occurrence,
        care_site,
    )
    hospital_name = CARE_SITE_LEVEL_NAMES["Hospital"]
    visit_predictor_by_level = {hospital_name: hospital_visit}

    if not hospital_only(care_site_levels=care_site_levels):
        visit_detail = prepare_visit_detail(data, start_date, end_date)

        uf_name = CARE_SITE_LEVEL_NAMES["UF"]
        uf_visit = get_uf_visit(
            visit_occurrence,
            visit_detail,
            care_site,
        )
        visit_predictor_by_level[uf_name] = uf_visit

        uc_name = CARE_SITE_LEVEL_NAMES["UC"]
        uc_visit = get_uc_visit(
            visit_occurrence,
            visit_detail,
            care_site,
        )
        visit_predictor_by_level[uc_name] = uc_visit

        uh_name = CARE_SITE_LEVEL_NAMES["UH"]
        uh_visit = get_uh_visit(
            visit_occurrence,
            visit_detail,
            care_site,
        )
        visit_predictor_by_level[uh_name] = uh_visit

        pole_name = CARE_SITE_LEVEL_NAMES["Pole"]
        pole_visit = get_pole_visit(
            uf_visit,
            care_site,
            care_site_relationship,
        )
        visit_predictor_by_level[pole_name] = pole_visit

    visit_predictor = concatenate_predictor_by_level(
        predictor_by_level=visit_predictor_by_level,
        care_site_levels=care_site_levels,
    )

    return compute_completeness(self, visit_predictor)


def compute_completeness(
    self,
    visit_predictor: DataFrame,
):
    partition_cols = [*self._index.copy(), "date"]

    n_visit = (
        visit_predictor.groupby(
            partition_cols,
            as_index=False,
            dropna=False,
        )
        .agg({"visit_id": "nunique"})
        .rename(columns={"visit_id": "n_visit"})
    )
    n_visit = to("pandas", n_visit)
    n_visit = impute_missing_dates(
        start_date=self.start_date,
        end_date=self.end_date,
        predictor=n_visit,
        partition_cols=partition_cols,
    )

    partition_cols = list(set(partition_cols) - {"date"})
    max_n_visit = (
        n_visit.groupby(
            partition_cols,
            as_index=False,
            dropna=False,
        )
        .agg({"n_visit": "max"})
        .rename(columns={"n_visit": "max_n_visit"})
    )

    visit_predictor = n_visit.merge(
        max_n_visit,
        on=partition_cols,
    )

    visit_predictor["c"] = visit_predictor["max_n_visit"].where(
        visit_predictor["max_n_visit"] == 0,
        visit_predictor["n_visit"] / visit_predictor["max_n_visit"],
    )
    return visit_predictor.drop(columns="max_n_visit")


def get_hospital_visit(
    visit_occurrence: DataFrame,
    care_site: DataFrame,
):
    hospital_visit = visit_occurrence.rename(
        columns={"visit_occurrence_id": "visit_id"}
    )

    hospital_visit = hospital_visit.merge(care_site, on="care_site_id")
    hospital_visit = hospital_visit[
        hospital_visit["care_site_level"] == CARE_SITE_LEVEL_NAMES["Hospital"]
    ]
    if is_koalas(hospital_visit):
        hospital_visit = hospital_visit.spark.cache()

    return hospital_visit


def get_uf_visit(
    visit_occurrence: DataFrame,
    visit_detail: DataFrame,
    care_site: DataFrame,
):
    uf_visit = visit_detail[visit_detail.visit_detail_type == VISIT_DETAIL_TYPE["UF"]]
    uf_visit = uf_visit.merge(
        visit_occurrence[
            visit_occurrence.columns.intersection(
                set(["visit_occurrence_id", "length_of_stay", "stay_type"])
            )
        ],
        on="visit_occurrence_id",
    ).drop(columns="visit_occurrence_id")
    uf_visit = uf_visit.merge(care_site, on="care_site_id")
    uf_visit = uf_visit[uf_visit["care_site_level"] == CARE_SITE_LEVEL_NAMES["UF"]]
    if is_koalas(uf_visit):
        uf_visit = uf_visit.spark.cache()

    return uf_visit


def get_uc_visit(
    visit_occurrence: DataFrame,
    visit_detail: DataFrame,
    care_site: DataFrame,
):
    uc_visit = visit_detail[visit_detail.visit_detail_type == VISIT_DETAIL_TYPE["UC"]]
    uc_visit = uc_visit.merge(
        visit_occurrence[
            visit_occurrence.columns.intersection(
                set(["visit_occurrence_id", "length_of_stay", "stay_type"])
            )
        ],
        on="visit_occurrence_id",
    ).drop(columns="visit_occurrence_id")
    uc_visit = uc_visit.merge(care_site, on="care_site_id")
    uc_visit = uc_visit[uc_visit["care_site_level"] == CARE_SITE_LEVEL_NAMES["UC"]]
    if is_koalas(uc_visit):
        uc_visit = uc_visit.spark.cache()

    return uc_visit


def get_uh_visit(
    visit_occurrence: DataFrame,
    visit_detail: DataFrame,
    care_site: DataFrame,
):
    uh_visit = visit_detail[visit_detail.visit_detail_type == VISIT_DETAIL_TYPE["UH"]]
    uh_visit = uh_visit.merge(
        visit_occurrence[
            visit_occurrence.columns.intersection(
                set(["visit_occurrence_id", "length_of_stay", "stay_type"])
            )
        ],
        on="visit_occurrence_id",
    ).drop(columns="visit_occurrence_id")
    uh_visit = uh_visit.merge(care_site, on="care_site_id")
    uh_visit = uh_visit[uh_visit["care_site_level"] == CARE_SITE_LEVEL_NAMES["UH"]]
    if is_koalas(uh_visit):
        uh_visit = uh_visit.spark.cache()

    return uh_visit


def get_pole_visit(
    uf_visit: DataFrame,
    care_site: DataFrame,
    care_site_relationship: DataFrame,
):
    care_site_cols = list(
        set(
            [
                "care_site_short_name",
                "care_site_level",
                "care_site_specialty",
                "specialties_set",
                "care_sites_set",
            ]
        ).intersection(uf_visit.columns)
    )
    pole_visit = convert_uf_to_pole(
        table=uf_visit.drop(columns=care_site_cols),
        table_name="uf_visit",
        care_site_relationship=care_site_relationship,
    )
    pole_visit = pole_visit.merge(care_site, on="care_site_id")
    pole_name = CARE_SITE_LEVEL_NAMES["Pole"]
    pole_visit = pole_visit[pole_visit["care_site_level"] == pole_name]
    if is_koalas(pole_visit):
        pole_visit = pole_visit.spark.cache()

    return pole_visit



import os
from pathlib import Path
from typing import List, Tuple

import pandas as pd

from . import settings


class LocalData:  # pragma: no cover
    """Pandas interface to OMOP data stored as local parquet files/folders.


    Parameters
    ----------
    folder: str
        absolute path to a folder containing several parquet files with omop data

    Examples
    --------
    >>> data = LocalData(folder="/export/home/USER/my_data/")
    >>> person = data.person
    >>> person.shape
    (100, 10)

    Attributes
    ----------
    person: pd.DataFrame
        Pandas dataframe `person`. All dataframe attributes are
        dynamically generated to match the content of the
        selected folder.

    """

    def __init__(
        self,
        folder: str,
    ):
        (
            self.available_tables,
            self.tables_paths,
            self.available_omop_tables,
        ) = self.list_available_tables(folder)
        if not self.available_omop_tables:
            raise ValueError(f"Folder {folder} does not contain any parquet omop data.")

    @staticmethod
    def list_available_tables(folder: str) -> Tuple[List[str], List[str]]:
        available_tables = []
        available_omop_tables = []
        tables_paths = {}
        known_omop_tables = settings.tables_to_load.keys()
        for filename in os.listdir(folder):
            file = Path(folder) / filename
            table_name = file.stem
            extension = file.suffix
            if extension == ".parquet":
                abspath = Path.resolve(file)
                tables_paths[table_name] = abspath
                available_tables.append(table_name)
                if table_name in known_omop_tables:
                    available_omop_tables.append(table_name)

        return available_tables, tables_paths, available_omop_tables

    def _read_table(self, table_name: str) -> pd.DataFrame:
        path = self.tables_paths[table_name]
        return pd.read_parquet(path)

    def __getattr__(self, table_name: str) -> pd.DataFrame:
        if table_name in self.available_tables:
            return self._read_table(table_name)
        raise AttributeError(f"Table '{table_name}' is not available in chosen folder.")

    def __dir__(self) -> List[str]:
        return list(super().__dir__()) + list(self.available_tables)



from typing import List

import altair as alt

vertical_bar_charts = dict(
    x=[
        {
            "title": "Stay type",
            "field": "stay_type",
            "type": "nominal",
            "sort": "-y",
        },
        {
            "title": "Length of stay",
            "field": "length_of_stay",
            "type": "nominal",
            "sort": "-y",
        },
    ],
    y=[
        dict(
            y=alt.Y(
                "sum(n_measurement):Q",
                title="Number of measurements",
                axis=alt.Axis(format="s"),
            ),
            tooltip=alt.Tooltip(
                "sum(n_measurement):Q",
                format=",",
            ),
            sort={
                "field": "n_measurement",
                "op": "sum",
                "order": "descending",
            },
        ),
    ],
)


def get_horizontal_bar_charts(standard_terminologies: List[str]):
    return dict(
        y=[
            {
                "title": "Care site",
                "field": "care_site_short_name",
                "sort": "-x",
            },
            {
                "title": "Care site specialty",
                "field": "care_site_specialty",
                "sort": "-x",
            },
            {
                "title": "Specialties-set",
                "field": "specialties_set",
                "sort": "-x",
            },
            {
                "title": "Care sites-set",
                "field": "care_sites_set",
                "sort": "-x",
            },
            {"title": "Concepts-set", "field": "concepts_set", "sort": "-x"},
        ]
        + [
            {
                "title": "{} concept".format(terminology),
                "field": "{}_concept_name".format(terminology),
                "sort": "-x",
            }
            for terminology in standard_terminologies
        ],
        x=[
            dict(
                x=alt.X(
                    "sum(n_measurement):Q",
                    title="Number of measurements",
                    axis=alt.Axis(format="s"),
                ),
                tooltip=alt.Tooltip(
                    "sum(n_measurement):Q",
                    format=",",
                ),
                sort={
                    "field": "n_measurement",
                    "op": "sum",
                    "order": "descending",
                },
            ),
        ],
    )


main_chart = dict(
    encode=dict(
        x=alt.X(
            "yearmonth(date):T",
            title="Time (Month Year)",
            axis=alt.Axis(tickCount="month", labelAngle=0, grid=True),
        ),
        y=alt.Y(
            "sum(n_measurement):Q",
            title="Number of measurements",
            axis=alt.Axis(grid=True),
        ),
        color=alt.Color(
            "value:N",
            sort={"field": "n_measurement", "op": "sum", "order": "descending"},
            title=None,
        ),
        tooltip=[
            alt.Tooltip("value:N", title="Index"),
            alt.Tooltip("yearmonth(date):T", title="Date"),
            alt.Tooltip(
                "sum(n_measurement):Q", title="Number of measurements", format=","
            ),
        ],
    ),
    properties=dict(
        height=300,
        width=900,
    ),
)

normalized_main_chart = dict(
    aggregates=[
        dict(
            sum_measurement="sum(n_measurement)",
            groupby=["value", "date"],
        ),
        dict(
            max_measurement="max(sum_measurement)",
            groupby=["value"],
        ),
    ],
    calculates=[
        dict(
            normalized_c=(alt.datum.sum_measurement / alt.datum.max_measurement)
            / alt.datum.c_0
        )
    ],
    legend_title="Mean",
    encode=dict(
        x=alt.X(
            "normalized_date:Q",
            title="Δt = (t - t₀) months",
            scale=alt.Scale(nice=False),
        ),
        y=alt.Y(
            "mean(normalized_c):Q",
            title="c(Δt) / c₀",
            axis=alt.Axis(grid=True),
        ),
        color=alt.Color(
            "value:N",
            sort={"field": "n_measurement", "op": "sum", "order": "descending"},
            title=None,
        ),
    ),
    properties=dict(
        height=300,
        width=900,
    ),
)

normalized_time_line = dict(
    encode=dict(
        x=alt.X(
            "normalized_date:Q",
            title="Δt = (t - t₀) months",
            scale=alt.Scale(nice=False),
        ),
        y=alt.Y(
            "sum(n_measurement):Q",
            title="Number of measurements",
            axis=alt.Axis(format="s"),
        ),
    ),
    properties=dict(
        height=50,
        width=900,
    ),
)

time_line = dict(
    encode=dict(
        x=alt.X(
            "yearmonth(date):T",
            title="Time (Month Year)",
            axis=alt.Axis(tickCount="month", labelAngle=0, grid=True),
        ),
        y=alt.Y(
            "sum(n_measurement):Q",
            title="Number of measurements",
            axis=alt.Axis(format="s"),
        ),
    ),
    properties=dict(
        height=50,
        width=900,
    ),
)

error_line = dict(
    legend_title="Standard deviation",
    mark_errorband=dict(
        extent="stdev",
    ),
    encode=dict(
        stroke=alt.Stroke(
            "legend_error_band",
            title="Error band",
            legend=alt.Legend(
                symbolType="square",
                orient="top",
                labelFontSize=12,
                labelFontStyle="bold",
            ),
        ),
    ),
)

chart_style = dict(
    labelFontSize=12,
    titleFontSize=13,
)



from typing import Callable, List

import pandas as pd

from edsteva.utils import loss_functions
from edsteva.utils.checks import check_columns


def error(
    predictor: pd.DataFrame,
    estimates: pd.DataFrame,
    index: List[str],
    loss_function: Callable = loss_functions.l2_loss,
    y: str = "c",
    y_0: str = "c_0",
    x: str = "date",
    name: str = "error",
):
    r"""Compute the error between the predictor $c(t)$ and the prediction $\hat{c}(t)$ as follow:

    $$
    error = \frac{\sum_{t_{min} \leq  t \leq t_{max}} \mathcal{l}(c(t), \hat{c}(t))}{t_{max} - t_{min}}
    $$

    Where the loss function $\mathcal{l}$ can be the L1 distance or the L2 distance.

    Parameters
    ----------
    predictor : pd.DataFrame
        $c(t)$ computed in the Probe
    estimates : pd.DataFrame
        $\hat{c}(t)$ computed in the Model
    index : List[str]
        Variable from which data is grouped
    loss_function : str, optional
        The loss function $\mathcal{l}$
    y : str, optional
        Target column name of $c(t)$
    y_0 : str, optional
        Target column name of $\hat{c}(t)$
    x : str, optional
        Target column name of $t$
    name : str, optional
        Column name of the output

    Example
    -------

    | care_site_level          | care_site_id | stay_type    | error |
    | :----------------------- | :----------- | :----------- | :---- |
    | Unité Fonctionnelle (UF) | 8312056386   | 'Urg_Hospit' | 0.040 |
    | Unité Fonctionnelle (UF) | 8312056386   | 'All'        | 0.028 |
    | Pôle/DMU                 | 8312027648   | 'Urg_Hospit' | 0.022 |
    | Pôle/DMU                 | 8312027648   | 'All'        | 0.014 |
    | Hôpital                  | 8312022130   | 'Urg_Hospit' | 0.027 |
    """
    check_columns(df=estimates, required_columns=[*index, y_0])
    check_columns(df=predictor, required_columns=[*index, x, y])

    fitted_predictor = predictor.merge(estimates, on=index)

    fitted_predictor["loss"] = loss_function(
        fitted_predictor[y] - fitted_predictor[y_0]
    )

    error = fitted_predictor.groupby(index)["loss"].mean().rename(name)

    return error.reset_index()



from edsteva.viz.plots.estimates_densities.estimates_densities import (
    estimates_densities_plot,
)
from edsteva.viz.plots.normalized_probe.normalized_probe import normalized_probe_plot
from edsteva.viz.plots.probe.wrapper import probe_plot

__all__ = [
    "probe_plot",
    "estimates_densities_plot",
    "normalized_probe_plot",
]



from typing import List

import pandas as pd

from edsteva.models import BaseModel
from edsteva.models.rectangle_function.algos import algos
from edsteva.models.rectangle_function.viz_configs import viz_configs


class RectangleFunction(BaseModel):
    r"""It models the completeness predictor $c(t)$ as a rectangle function $f_{t_0, c_0, t_1}(t)$ as follow:

    $$
    f_{t_0, c_0, t_1}(t) = c_0 \ \mathbb{1}_{t_0 \leq t \leq t_1}(t)
    $$

    It computes the following estimates $(t_0, c_0, t_1)$:

    - the characteristic time $t_0$ estimates the time after which the data is available.
    - the characteristic time $t_1$ estimates the time after which the data is not available anymore.
    - the characteristic value $c_0$ estimates the completeness between $t_0$ and $t_1$.

    Attributes
    ----------
    _algo: List[str]
        Algorithm used to compute the estimates
        **VALUE**: ``"loss_minimization"``
    _coefs: List[str]
        Model coefficients
        **VALUE**: ``["t_0", "c_0", "t_1"]``
    _default_metrics: List[str]
        Metrics to used by default
        **VALUE**: ``[error_between_t0_t1]``
    _viz_config: List[str]
        Dictionary of configuration for visualization purpose.
        **VALUE**: ``{}``

    Example
    ----------

    ```python
    from edsteva.models.step_function import StepFunction

    step_function_model = StepFunction()
    step_function_model.fit(probe)
    step_function_model.estimates.head()
    ```

    | care_site_level          | care_site_id | stay_type | t_0        | c_0   | t_1        | error |
    | :----------------------- | :----------- | :-------- | :--------- | :---- | :--------- | :---- |
    | Unité Fonctionnelle (UF) | 8312056386   | 'Urg'     | 2019-05-01 | 0.397 | 2020-05-01 | 0.040 |
    | Unité Fonctionnelle (UF) | 8312056386   | 'All'     | 2011-04-01 | 0.583 | 2013-04-01 | 0.028 |
    | Pôle/DMU                 | 8312027648   | 'Hospit'  | 2021-03-01 | 0.677 | 2022-03-01 | 0.022 |
    | Pôle/DMU                 | 8312027648   | 'All'     | 2018-08-01 | 0.764 | 2019-08-01 | 0.014 |
    | Hôpital                  | 8312022130   | 'Hospit'  | 2022-02-01 | 0.652 | 2022-08-01 | 0.027 |
    """

    def __init__(
        self,
        algo: str = "loss_minimization",
    ):
        """Initialisation of the RectangleFunction Model.

        Parameters
        ----------
        algo : Callable, optional
            Algorithm used for the coefficients estimation ($t_0$, $t_1$ and $c_0$)
        """
        coefs = ["t_0", "c_0", "t_1"]
        default_metrics = ["error_between_t0_t1"]
        super().__init__(
            algo=algo,
            coefs=coefs,
            default_metrics=default_metrics,
        )

    def fit_process(
        self,
        predictor: pd.DataFrame,
        index: List[str] = None,
        **kwargs,
    ):
        """Script to be used by [``fit()``][edsteva.models.base.BaseModel.fit]

        Parameters
        ----------
        predictor : pd.DataFrame
            Target variable to be fitted
        index : List[str], optional
            Variable from which data is grouped
            **EXAMPLE**: `["care_site_level", "stay_type", "note_type", "care_site_id"]`
        """
        return algos.get(self._algo)(predictor=predictor, index=index, **kwargs)

    def predict_process(
        self,
        predictor: pd.DataFrame,
        index: List[str],
    ):
        """Script to be used by [``predict()``][edsteva.models.base.BaseModel.predict]

        Parameters
        ----------
        predictor : pd.DataFrame
            Target DataFrame to be predicted
        index : List[str]
            List of the columns given by Probe._index

        Returns
        -------
        pd.DataFrame
            Prediction

        Raises
        ------
        Exception
            Some indexes have no associated estimates, the model must be fitted on an adequate probe

        Examples
        --------
        | care_site_level          | care_site_id | care_site_short_name | stay_type    | date       | n_visit | c     | c_fit |
        | :----------------------- | :----------- | :------------------- | :----------- | :--------- | :------ | :---- | :---- |
        | Unité Fonctionnelle (UF) | 8312056386   | Care site 1          | 'Urg_Hospit' | 2019-05-01 | 233.0   | 0.841 | 0.758 |
        | Unité Fonctionnelle (UF) | 8312056386   | Care site 1          | 'All'        | 2021-04-01 | 393.0   | 0.640 | 0.758 |
        | Pôle/DMU                 | 8312027648   | Care site 2          | 'Urg_Hospit' | 2011-03-01 | 204.0   | 0.497 | 0     |
        | Pôle/DMU                 | 8312027648   | Care site 2          | 'All'        | 2018-08-01 | 22.0    | 0.784 | 0.874 |
        | Hôpital                  | 8312022130   | Care site 3          | 'Urg_Hospit' | 2022-02-01 | 9746.0  | 0.974 | 0.912 |
        """
        prediction = self.is_predictable_probe(predictor, index)

        rect_mask = (prediction["date"] >= prediction["t_0"]) & (
            prediction["date"] <= prediction["t_1"]
        )
        prediction["c_hat"] = prediction["c_0"].where(rect_mask, 0)

        return prediction.drop(columns=self._metrics)

    def get_viz_config(self, viz_type: str, **kwargs):
        if viz_type in viz_configs.keys():
            _viz_config = self._viz_config.get(viz_type)
            if _viz_config is None:
                _viz_config = "default"
        else:
            raise ValueError(f"edsteva has no {viz_type} registry !")
        return viz_configs[viz_type].get(_viz_config)(self, **kwargs)



from edsteva.viz.dashboards.normalized_probe.normalized_probe import (
    normalized_probe_dashboard,
)
from edsteva.viz.dashboards.probe.wrapper import probe_dashboard

__all__ = [
    "probe_dashboard",
    "normalized_probe_dashboard",
]



from .defaults import (
    chart_style,
    error_line,
    horizontal_bar_charts,
    normalized_main_chart,
    normalized_time_line,
    vertical_bar_charts,
)


def get_normalized_probe_dashboard_config(self):
    return dict(
        chart_style=chart_style,
        main_chart=normalized_main_chart,
        time_line=normalized_time_line,
        error_line=error_line,
        vertical_bar_charts=vertical_bar_charts,
        horizontal_bar_charts=horizontal_bar_charts,
    )



from pathlib import Path

import _pickle as pickle
from loguru import logger


def save_object(obj, filename: Path):
    if not isinstance(filename, Path):
        filename = Path(filename)
    Path.mkdir(filename.parent, exist_ok=True, parents=True)
    with Path.open(filename, "wb") as outp:  # Overwrites any existing file.
        pickle.dump(obj, outp, -1)
        logger.info("Saved to {}", filename)


def load_object(filename: str):
    file = Path(filename)
    if Path.is_file(file):
        with Path.open(file, "rb") as obj:
            logger.info("Successfully loaded from {}", filename)
            return pickle.load(obj)
    else:
        raise FileNotFoundError(
            "There is no file found in {}".format(
                filename,
            )
        )


def delete_object(obj, filename: str):
    file = Path(filename)
    if Path.is_file(file):
        Path.unlink(file)
        logger.info(
            "Removed from {}",
            filename,
        )
        if hasattr(obj, "path"):
            del obj.path
    else:
        logger.warning(
            "There is no file found in {}",
            filename,
        )



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['edsteva',
 'edsteva.io',
 'edsteva.io.synthetic',
 'edsteva.metrics',
 'edsteva.models',
 'edsteva.models.rectangle_function',
 'edsteva.models.rectangle_function.algos',
 'edsteva.models.rectangle_function.viz_configs',
 'edsteva.models.step_function',
 'edsteva.models.step_function.algos',
 'edsteva.models.step_function.viz_configs',
 'edsteva.probes',
 'edsteva.probes.biology',
 'edsteva.probes.biology.completeness_predictors',
 'edsteva.probes.biology.viz_configs',
 'edsteva.probes.biology.viz_configs.n_measurement',
 'edsteva.probes.biology.viz_configs.per_measurement',
 'edsteva.probes.biology.viz_configs.per_visit',
 'edsteva.probes.condition',
 'edsteva.probes.condition.completeness_predictors',
 'edsteva.probes.condition.viz_configs',
 'edsteva.probes.condition.viz_configs.n_condition',
 'edsteva.probes.condition.viz_configs.per_condition',
 'edsteva.probes.condition.viz_configs.per_visit',
 'edsteva.probes.note',
 'edsteva.probes.note.completeness_predictors',
 'edsteva.probes.note.viz_configs',
 'edsteva.probes.note.viz_configs.n_note',
 'edsteva.probes.note.viz_configs.per_note',
 'edsteva.probes.note.viz_configs.per_visit',
 'edsteva.probes.utils',
 'edsteva.probes.visit',
 'edsteva.probes.visit.completeness_predictors',
 'edsteva.probes.visit.viz_configs',
 'edsteva.probes.visit.viz_configs.n_visit',
 'edsteva.probes.visit.viz_configs.per_visit',
 'edsteva.utils',
 'edsteva.viz',
 'edsteva.viz.dashboards',
 'edsteva.viz.dashboards.normalized_probe',
 'edsteva.viz.dashboards.probe',
 'edsteva.viz.plots',
 'edsteva.viz.plots.estimates_densities',
 'edsteva.viz.plots.normalized_probe',
 'edsteva.viz.plots.probe']

package_data = \
{'': ['*']}

install_requires = \
['altair>=5.0,<6.0',
 'catalogue>=2.0.8,<3.0.0',
 'ipython>=7.31.0,<8.0.0',
 'koalas>=1.8.2,<2.0.0',
 'loguru==0.7.0',
 'numpy<1.20.0',
 'pandas>=1.3,<2.0',
 'pgpasslib>=1.1.0,<2.0.0',
 'psycopg2-binary>=2.9.3,<3.0.0',
 'pyarrow>=0.15,<0.17.0',
 'pyspark>=2.4.3,<2.5.0']

setup_kwargs = {
    'name': 'edsteva',
    'version': '0.2.2',
    'description': 'EDS-TeVa provides a set of tools that aims at modeling the adoption over time and across space of the Electronic Health Records.',
    'long_description': '<p align="center">\n<b>DISCLAIMER: </b>EDS-TeVa is intended to be a module of <a href="https://github.com/aphp/EDS-Scikit">EDS-Scikit</a>\n</p>\n\n<div align="center">\n\n<p align="center">\n  <a href="https://aphp.github.io/edsteva/latest/"><img src="https://aphp.github.io/edsteva/latest/assets/logo/edsteva_logo_small.svg" alt="EDS-TeVa"></a>\n</p>\n\n# EDS-TeVa [![Binder](https://mybinder.org/badge_logo.svg)](https://mybinder.org/v2/gh/aphp/edsteva/HEAD?labpath=notebooks%2Fsynthetic_data.ipynb)\n\n<p align="center">\n<a href="https://aphp.github.io/edsteva/latest/" target="_blank">\n    <img src="https://img.shields.io/github/actions/workflow/status/aphp/edsteva/documentation.yaml?branch=main&label=docs&style=flat" alt="Documentation">\n</a>\n<a href="https://pypi.org/project/edsteva/" target="_blank">\n    <img src="https://img.shields.io/pypi/v/edsteva?color=blue&style=flat" alt="PyPI">\n</a>\n<a href="https://codecov.io/github/aphp/edsteva?branch=main" target="_blank">\n    <img src="https://codecov.io/github/aphp/edsteva/coverage.svg?branch=main" alt="Codecov">\n</a>\n<a href="https://github.com/psf/black" target="_blank">\n    <img src="https://img.shields.io/badge/code%20style-black-000000.svg" alt="Black">\n</a>\n<a href="https://python-poetry.org/" target="_blank">\n    <img src="https://img.shields.io/badge/reproducibility-poetry-blue" alt="Poetry">\n</a>\n<a href="https://www.python.org/" target="_blank">\n    <img src="https://img.shields.io/badge/python-~3.7.1-brightgreen" alt="Supported Python versions">\n</a>\n<a href="https://github.com/astral-sh/ruff" target="_blank">\n    <img src="https://img.shields.io/endpoint?url=https://raw.githubusercontent.com/charliermarsh/ruff/main/assets/badge/v2.json" alt="Ruff">\n</a>\n</p>\n</div>\n\n**Documentation**: <a href="https://aphp.github.io/edsteva/latest/" target="_blank">https://aphp.github.io/edsteva/latest/</a>\n\n**Source Code**: <a href="https://github.com/aphp/edsteva" target="_blank">https://github.com/aphp/edsteva</a>\n\n---\n\nEDS-TeVa provides a set of tools that aims at modeling the adoption over time and across space of the Electronic Health Records.\n\n## Requirements\nEDS-TeVa stands on the shoulders of [Spark 2.4](https://spark.apache.org/docs/2.4.8/index.html) which requires:\n\n- Python ~3.7.1\n- Java 8\n\n## Installation\n\nYou can install EDS-TeVa through ``pip``:\n\n```shell\npip install edsteva\n```\nWe recommend pinning the library version in your projects, or use a strict package manager like [Poetry](https://python-poetry.org/).\n\n```\npip install edsteva==0.2.2\n```\n## Example\n\nA scientific paper is currently being written that describes extensively the use of the library on the study of pulmonary embolism of cancer patients.\n\n## Contributing\n\nContributions are welcome, and they are greatly appreciated! Every little bit helps, and credit will always be given.\n\n## Acknowledgement\n\nWe would like to thank [Assistance Publique – Hôpitaux de Paris](https://www.aphp.fr/) and [AP-HP Foundation](https://fondationrechercheaphp.fr/) for funding this project.\n',
    'author': 'Adam Remaki',
    'author_email': 'adam.remaki@aphp.fr',
    'maintainer': 'Adam Remaki',
    'maintainer_email': 'adam.remaki@aphp.fr',
    'url': 'https://github.com/aphp/edsteva',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7.1,<3.8.0',
}


setup(**setup_kwargs)



import numpy as np
import pandas as pd


def generate_events_before_t0(
    generator: np.random.Generator,
    t_start: int,
    t_end: int,
    n_events: int,
    t0: int,
    increase_time: int,
    increase_ratio: float,
):
    t0_before = t0 - increase_time / 2
    n_before = int(
        (t0_before - t_start)
        * n_events
        / ((t0 - t_start) + increase_ratio * (t_end - t0))
    )

    return pd.to_datetime(
        pd.Series(generator.integers(t_start, t0_before, n_before)),
        unit="s",
    )


def generate_events_after_t0(
    generator: np.random.Generator,
    t_start: int,
    t_end: int,
    n_events: int,
    t0: int,
    increase_time: int,
    increase_ratio: float,
):
    t0_after = t0 + increase_time / 2
    n_after = int(
        increase_ratio
        * (t_end - t0_after)
        * n_events
        / ((t0 - t_start) + increase_ratio * (t_end - t0))
    )

    return pd.to_datetime(
        pd.Series(generator.integers(t0_after, t_end, n_after)),
        unit="s",
    )


def generate_events_around_t0(
    generator: np.random.Generator,
    t_start: int,
    t_end: int,
    n_events: int,
    t0: int,
    increase_time: int,
    increase_ratio: float,
):
    t0_before = t0 - increase_time / 2
    t0_after = t0 + increase_time / 2
    n_middle = int(
        (increase_time / 2)
        * (increase_ratio + 1)
        * n_events
        / ((t0 - t_start) + increase_ratio * (t_end - t0))
    )

    return pd.to_datetime(
        pd.Series(
            generator.triangular(
                left=t0_before, right=t0_after, mode=t0_after, size=n_middle
            )
        ),
        unit="s",
    )


def generate_events_around_t1(
    generator: np.random.Generator,
    t_start: int,
    t_end: int,
    n_events: int,
    t1: int,
    increase_time: int,
    increase_ratio: float,
):
    t1_before = t1 - increase_time / 2
    t1_after = t1 + increase_time / 2
    n_middle = int(
        (increase_time / 2)
        * (increase_ratio + 1)
        * n_events
        / ((t1 - t_start) * increase_ratio + (t_end - t1))
    )

    return pd.to_datetime(
        pd.Series(
            generator.triangular(
                left=t1_before, right=t1_after, mode=t1_before, size=n_middle
            )
        ),
        unit="s",
    )


def generate_events_after_t1(
    generator: np.random.Generator,
    t_start: int,
    t_end: int,
    n_events: int,
    t1: int,
    increase_time: int,
    increase_ratio: float,
):
    t1_after = t1 + increase_time / 2
    n_after = int(
        (t_end - t1_after) * n_events / ((t1 - t_start) * increase_ratio + (t_end - t1))
    )

    return pd.to_datetime(
        pd.Series(generator.integers(t1_after, t_end, n_after)),
        unit="s",
    )


def recursive_items(dictionary):
    for key, value in dictionary.items():
        if type(value) is dict:
            if key[0] != "P":
                yield key
            yield from recursive_items(value)
        else:
            yield key



__version__ = "0.2.2"


import importlib
import os
import sys
import time
from distutils.version import LooseVersion
from pathlib import Path
from typing import List, Tuple

import pyarrow
import pyspark
from loguru import logger
from pyspark import SparkContext
from pyspark.sql import SparkSession

logger.remove()
logger.add(sys.stderr, level="INFO")


def koalas_options() -> None:
    """
    Set necessary options to optimise Koalas
    """

    # Reloading Koalas to use the new configuration
    ks = sys.modules.get("databricks.koalas", None)

    if ks is not None:
        importlib.reload(ks)

    else:  # pragma: no cover
        import databricks.koalas as ks

    ks.set_option("compute.default_index_type", "distributed")
    ks.set_option("compute.ops_on_diff_frames", True)


def set_env_variables() -> None:
    # From https://github.com/databricks/koalas/blob/master/databricks/koalas/__init__.py
    if LooseVersion(pyspark.__version__) < LooseVersion("3.0"):
        if LooseVersion(pyarrow.__version__) >= LooseVersion("0.15"):
            os.environ["ARROW_PRE_0_15_IPC_FORMAT"] = "1"

    if LooseVersion(pyarrow.__version__) >= LooseVersion("2.0.0"):  # pragma: no cover
        os.environ["PYARROW_IGNORE_TIMEZONE"] = "0"


def improve_performances(
    to_add_conf: List[Tuple[str, str]] = None,
    quiet_spark: bool = True,
) -> Tuple[SparkSession, SparkContext, SparkSession.sql]:
    """
    (Re)defines various Spark variable with some configuration changes
    to improve performances by enabling Arrow
    This has to be done
    - Before launching a SparkCOntext
    - Before importing Koalas
    Those two points are being taken care on this function.
    If a SparkSession already exists, it will copy its configuration before
    creating a new one

    Returns
    -------
    Tuple of
    - A SparkSession
    - The associated SparkContext
    - The associated ``sql`` object to run SQL queries
    """

    # Check if a spark Session is up
    global spark, sc, sql

    spark = SparkSession.builder.getOrCreate()
    sc = spark.sparkContext

    if quiet_spark:
        sc.setLogLevel("ERROR")

    conf = sc.getConf()

    # Synchronizing TimeZone
    tz = os.environ.get("TZ", "UTC")
    os.environ["TZ"] = tz
    time.tzset()

    if to_add_conf is None:
        to_add_conf = []

    to_add_conf.extend(
        [
            ("spark.app.name", f"{os.environ.get('USER')}_scikit"),
            ("spark.sql.session.timeZone", tz),
            ("spark.sql.execution.arrow.enabled", "true"),
            ("spark.sql.execution.arrow.pyspark.enabled", "true"),
        ]
    )

    for key, value in to_add_conf:
        conf.set(key, value)

    # Stopping context to add necessary env variables
    sc.stop()
    spark.stop()

    set_env_variables()

    spark = SparkSession.builder.enableHiveSupport().config(conf=conf).getOrCreate()

    sc = spark.sparkContext

    if quiet_spark:
        sc.setLogLevel("ERROR")

    sql = spark.sql

    koalas_options()

    return spark, sc, sql


CACHE_DIR = Path.home() / ".cache" / "edsteva"



from datetime import datetime
from typing import Dict, List

import pandas as pd
from loguru import logger

from edsteva.utils.framework import get_framework
from edsteva.utils.typing import DataFrame

CARE_SITE_LEVEL_NAMES = {
    "Hospital": "Hôpital",
    "Pole": "Pôle/DMU",
    "UF": "Unité Fonctionnelle (UF)",
    "UC": "Unité de consultation (UC)",
    "UH": "Unité d’hébergement (UH)",
}

VISIT_DETAIL_TYPE = {
    "UF": "PASS UF",
    "UC": "PASS UC",
    "UH": "PASS UH",
}


def impute_missing_dates(
    start_date: datetime,
    end_date: datetime,
    predictor: DataFrame,
    partition_cols: List[str],
):
    # Generate all available dates
    if not start_date:
        start_date = predictor["date"].min()
    if not end_date:
        end_date = predictor["date"].max()
    date_index = pd.date_range(
        start=start_date,
        end=end_date,
        freq="MS",
        closed="left",
    )
    date_index = pd.DataFrame({"date": date_index})

    # Generate all available partitions
    all_partitions = (
        predictor[list(set(partition_cols) - {"date"})]
        .drop_duplicates()
        .merge(date_index, how="cross")
    )
    return all_partitions.merge(
        predictor,
        on=partition_cols,
        how="left",
    ).fillna({col: 0 for col in set(predictor.columns) - set(partition_cols)})


def hospital_only(care_site_levels: List[str]):
    if not isinstance(care_site_levels, list):
        care_site_levels = [care_site_levels]
    return len(care_site_levels) == 1 and (
        care_site_levels[0] == "Hospital"
        or care_site_levels[0] == CARE_SITE_LEVEL_NAMES["Hospital"]
    )


def concatenate_predictor_by_level(
    predictor_by_level: Dict[str, DataFrame],
    care_site_levels: List[str] = None,
) -> DataFrame:
    predictors_to_concat = []
    if care_site_levels:
        if not isinstance(care_site_levels, list):
            care_site_levels = [care_site_levels]
        unknown_levels, unavailable_levels, selected_levels = [], [], []
        for level in care_site_levels:
            if level in predictor_by_level.keys():
                predictors_to_concat.append(predictor_by_level[level])
                selected_levels.append(level)
            elif level in CARE_SITE_LEVEL_NAMES.keys():
                raw_level = CARE_SITE_LEVEL_NAMES[level]
                if raw_level in predictor_by_level.keys():
                    predictors_to_concat.append(predictor_by_level[raw_level])
                    selected_levels.append(level)
                else:
                    unavailable_levels.append(level)
            else:
                unknown_levels.append(level)

        logger.debug(
            "The following levels {} have been selected",
            selected_levels,
        )
        if unknown_levels:
            logger.warning(
                "Unrecognized levels {}.the only supported levels are: {}",
                unknown_levels,
                list(CARE_SITE_LEVEL_NAMES.values())
                + list(CARE_SITE_LEVEL_NAMES.keys()),
            )
        if unavailable_levels:
            logger.warning(
                "The following levels: {} are not available for this probe.",
                unavailable_levels,
            )
    else:
        predictors_to_concat = list(predictor_by_level.values())
        logger.debug(
            "The following levels {} have been selected",
            list(predictor_by_level.keys()),
        )

    if not predictors_to_concat:
        raise AttributeError(
            "care site levels must include at least one of the following levels: {}".format(
                list(CARE_SITE_LEVEL_NAMES.values())
                + list(CARE_SITE_LEVEL_NAMES.keys())
            )
        )

    return get_framework(predictors_to_concat[0]).concat(predictors_to_concat)


def get_child_and_parent_cs(
    care_site_sample: DataFrame, care_site_relationship: DataFrame
):
    extended_care_site_id_to_filter = []

    # Parent care site to get
    parent_rel = care_site_relationship[
        ~care_site_relationship.parent_care_site_id.isna()
    ][
        [
            "care_site_id",
            "parent_care_site_id",
            "parent_care_site_level",
            "parent_care_site_specialty",
        ]
    ]
    parent_care_site_filter = care_site_sample.copy()
    while set(parent_care_site_filter.care_site_level.unique()).intersection(
        CARE_SITE_LEVEL_NAMES.values()
    ):
        extended_care_site_id_to_filter.append(
            parent_care_site_filter[
                ["care_site_id", "care_site_level", "care_site_specialty"]
            ]
        )
        parent_care_site_filter = parent_care_site_filter.merge(
            parent_rel,
            on="care_site_id",
        )[
            [
                "parent_care_site_id",
                "parent_care_site_level",
                "parent_care_site_specialty",
            ]
        ].rename(
            columns={
                "parent_care_site_id": "care_site_id",
                "parent_care_site_level": "care_site_level",
                "parent_care_site_specialty": "care_site_specialty",
            }
        )

    # Child care site to get
    child_rel = care_site_relationship[~care_site_relationship.care_site_id.isna()][
        [
            "care_site_id",
            "care_site_level",
            "care_site_specialty",
            "parent_care_site_id",
        ]
    ].rename(
        columns={
            "care_site_id": "child_care_site_id",
            "care_site_level": "child_care_site_level",
            "care_site_specialty": "child_care_site_specialty",
            "parent_care_site_id": "care_site_id",
        }
    )
    child_care_site_filter = care_site_sample.copy()
    while set(child_care_site_filter.care_site_level.unique()).intersection(
        CARE_SITE_LEVEL_NAMES.values()
    ):
        extended_care_site_id_to_filter.append(
            child_care_site_filter[
                ["care_site_id", "care_site_level", "care_site_specialty"]
            ]
        )
        child_care_site_filter = child_care_site_filter.merge(
            child_rel,
            on="care_site_id",
        )[
            [
                "child_care_site_id",
                "child_care_site_level",
                "child_care_site_specialty",
            ]
        ].rename(
            columns={
                "child_care_site_id": "care_site_id",
                "child_care_site_level": "care_site_level",
                "child_care_site_specialty": "care_site_specialty",
            }
        )

    return pd.concat(extended_care_site_id_to_filter).drop_duplicates()



from datetime import datetime
from typing import Dict, List, Tuple, Union

import pandas as pd
from loguru import logger

from edsteva.probes.utils.prepare_df import (
    prepare_biology_relationship,
    prepare_care_site,
    prepare_measurement,
    prepare_visit_occurrence,
)
from edsteva.probes.utils.utils import (
    CARE_SITE_LEVEL_NAMES,
    concatenate_predictor_by_level,
    hospital_only,
    impute_missing_dates,
)
from edsteva.utils.checks import check_tables
from edsteva.utils.framework import is_koalas, to
from edsteva.utils.typing import Data, DataFrame


def compute_completeness_predictor_per_measurement(
    self,
    data: Data,
    care_site_relationship: pd.DataFrame,
    start_date: datetime,
    end_date: datetime,
    care_site_levels: List[str],
    stay_types: Union[str, Dict[str, str]],
    care_site_ids: List[int],
    care_site_short_names: List[str],
    care_site_specialties: List[str],
    concept_codes: List[str],
    care_sites_sets: Union[str, Dict[str, str]],
    specialties_sets: Union[str, Dict[str, str]],
    concepts_sets: Union[str, Dict[str, str]],
    length_of_stays: List[float],
    source_terminologies: Dict[str, str],
    mapping: List[Tuple[str, str, str]],
    **kwargs
):
    r"""Script to be used by [``compute()``][edsteva.probes.base.BaseProbe.compute]

    The ``per_measurement`` algorithm computes $c_(t)$ the availability of biological measurements:

    $$
    c(t) = \frac{n_{biology}(t)}{n_{max}}
    $$

    Where $n_{biology}(t)$ is the number of biological measurements, $t$ is the month and $n_{max} = \max_{t}(n_{biology}(t))$.
    """
    self._metrics = ["c", "n_measurement"]
    check_tables(
        data=data,
        required_tables=["measurement", "concept", "concept_relationship"],
    )
    standard_terminologies = self._standard_terminologies
    biology_relationship = prepare_biology_relationship(
        data=data,
        standard_terminologies=standard_terminologies,
        source_terminologies=source_terminologies,
        mapping=mapping,
    )

    self.biology_relationship = biology_relationship
    root_terminology = mapping[0][0]

    measurement = prepare_measurement(
        data=data,
        biology_relationship=biology_relationship,
        concept_codes=concept_codes,
        concepts_sets=concepts_sets,
        start_date=start_date,
        end_date=end_date,
        root_terminology=root_terminology,
        standard_terminologies=standard_terminologies,
        per_visit=False,
    )

    visit_occurrence = prepare_visit_occurrence(
        data=data,
        start_date=None,
        end_date=None,
        stay_types=stay_types,
        length_of_stays=length_of_stays,
    ).drop(columns=["visit_occurrence_source_value", "date"])

    care_site = prepare_care_site(
        data=data,
        care_site_ids=care_site_ids,
        care_site_short_names=care_site_short_names,
        care_site_specialties=care_site_specialties,
        care_sites_sets=care_sites_sets,
        specialties_sets=specialties_sets,
        care_site_relationship=care_site_relationship,
    )

    hospital_visit = get_hospital_measurements(
        measurement=measurement,
        visit_occurrence=visit_occurrence,
        care_site=care_site,
    )
    hospital_name = CARE_SITE_LEVEL_NAMES["Hospital"]
    biology_predictor_by_level = {hospital_name: hospital_visit}

    if care_site_levels and not hospital_only(care_site_levels=care_site_levels):
        logger.info(
            "Biological measurements are only available at hospital level for now"
        )
        care_site_levels = "Hospital"

    biology_predictor = concatenate_predictor_by_level(
        predictor_by_level=biology_predictor_by_level,
        care_site_levels=care_site_levels,
    )

    return compute_completeness(self, biology_predictor)


def compute_completeness(
    self,
    biology_predictor: DataFrame,
):
    partition_cols = [*self._index.copy(), "date"]
    n_measurement = (
        biology_predictor.groupby(
            partition_cols,
            as_index=False,
            dropna=False,
        )
        .agg({"measurement_id": "nunique"})
        .rename(columns={"measurement_id": "n_measurement"})
    )
    n_measurement = to("pandas", n_measurement)
    n_measurement = impute_missing_dates(
        start_date=self.start_date,
        end_date=self.end_date,
        predictor=n_measurement,
        partition_cols=partition_cols,
    )

    partition_cols = list(set(partition_cols) - {"date"})
    max_measurement = (
        n_measurement.groupby(
            partition_cols,
            as_index=False,
            dropna=False,
        )[["n_measurement"]]
        .agg({"n_measurement": "max"})
        .rename(columns={"n_measurement": "max_n_measurement"})
    )

    biology_predictor = n_measurement.merge(
        max_measurement,
        on=partition_cols,
    )

    biology_predictor["c"] = biology_predictor["max_n_measurement"].where(
        biology_predictor["max_n_measurement"] == 0,
        biology_predictor["n_measurement"] / biology_predictor["max_n_measurement"],
    )
    return biology_predictor.drop(columns="max_n_measurement")


def get_hospital_measurements(
    measurement: DataFrame,
    visit_occurrence: DataFrame,
    care_site: DataFrame,
):
    hospital_measurement = measurement.merge(
        visit_occurrence,
        on="visit_occurrence_id",
    )
    hospital_measurement = hospital_measurement.merge(care_site, on="care_site_id")

    if is_koalas(hospital_measurement):
        hospital_measurement = hospital_measurement.spark.cache()

    return hospital_measurement



from datetime import datetime
from functools import reduce
from math import ceil, floor, log10
from pathlib import Path
from typing import Dict, List, Union

import altair as alt
import pandas as pd
from loguru import logger

from edsteva.probes.utils.filter_df import filter_table_by_date
from edsteva.probes.utils.utils import CARE_SITE_LEVEL_NAMES


def generate_main_chart(
    base: alt.Chart,
    main_chart_config: Dict[str, str],
    index_selection: alt.SelectionParameter = None,
    index_fields: List[str] = None,
    x_axis_title: str = None,
    y_axis_title: str = None,
):
    if x_axis_title:
        main_chart_config["encode"]["x"]["title"] = x_axis_title
    if y_axis_title:
        main_chart_config["encode"]["y"]["title"] = y_axis_title
    if index_fields:
        base = base.transform_fold(index_fields, as_=["index", "value"])
        if "aggregates" in main_chart_config.keys():
            for aggregate in main_chart_config["aggregates"]:
                base = base.transform_joinaggregate(**aggregate)
        if "calculates" in main_chart_config.keys():
            for calculate in main_chart_config["calculates"]:
                base = base.transform_calculate(**calculate)
        if "filters" in main_chart_config.keys():
            for filter in main_chart_config["filters"]:
                base = base.transform_filter(**filter)
        main_chart = base.encode(**main_chart_config["encode"])
        if index_selection:
            main_chart = main_chart.transform_filter(index_selection)
    else:
        main_chart = base.encode(
            x=main_chart_config["encode"]["x"],
            y=main_chart_config["encode"]["y"],
        )

    return main_chart.properties(**main_chart_config["properties"])


def generate_model_line(
    main_chart: alt.Chart,
    model_line_config: Dict[str, str],
):
    model_line = main_chart.mark_line(**model_line_config["mark_line"])
    if "aggregates" in model_line_config.keys():
        for aggregate in model_line_config["aggregates"]:
            model_line = model_line.transform_joinaggregate(**aggregate)
    if "calculates" in model_line_config.keys():
        for calculate in model_line_config["calculates"]:
            model_line = model_line.transform_calculate(**calculate)
    if "filters" in model_line_config.keys():
        for filter in model_line_config["filters"]:
            model_line = model_line.transform_filter(**filter)
    model_line = model_line.encode(**model_line_config["encode"])
    return add_selection_on_legend(model_line)


def generate_error_line(
    main_chart: alt.Chart,
    error_line_config: Dict[str, str],
):
    error_line = main_chart.mark_errorband(
        **error_line_config["mark_errorband"]
    ).encode(**error_line_config["encode"])
    return add_selection_on_legend(error_line, opacity_true=0.3, opacity_false=0.05)


def generate_probe_line(
    main_chart: alt.Chart,
    probe_line_config: Dict[str, str],
):
    probe_line = main_chart.mark_line().encode(**probe_line_config["encode"])
    return add_selection_on_legend(probe_line)


def generate_time_line(
    base: alt.Chart,
    time_line_config: Dict[str, str],
):
    time_selection = alt.selection_interval(encodings=["x"])
    time_line = (
        base.mark_line().encode(**time_line_config["encode"]).add_params(time_selection)
    ).properties(**time_line_config["properties"])
    return time_line, time_selection


def generate_horizontal_bar_charts(
    base: alt.Chart,
    horizontal_bar_charts_config: Dict[str, str],
    predictor: pd.DataFrame,
):
    horizontal_bar_charts = {}
    y_variables_selections = {}
    for y_variable in horizontal_bar_charts_config["y"]:
        if y_variable["field"] not in predictor.columns:
            continue
        y_variable_bar_charts = []
        y_variable_selection = alt.selection_point(fields=[y_variable["field"]])
        y_variables_selections[y_variable["field"]] = y_variable_selection
        y_variable_base_chart = (
            base.mark_bar()
            .encode(
                y=alt.Y(**y_variable),
            )
            .add_params(y_variable_selection)
        )
        for x_variable in horizontal_bar_charts_config["x"]:
            y_index_variable_color = alt.condition(
                y_variables_selections[y_variable["field"]],
                alt.Color(
                    "{}:N".format(y_variable["field"]),
                    legend=None,
                    sort=x_variable["sort"],
                ),
                alt.value("lightgray"),
            )
            y_variable_bar_chart = y_variable_base_chart.encode(
                x=x_variable["x"],
                color=y_index_variable_color,
                tooltip=x_variable["tooltip"],
            )

            y_variable_bar_charts.append(y_variable_bar_chart)
        horizontal_bar_charts[y_variable["field"]] = y_variable_bar_charts
    return horizontal_bar_charts, y_variables_selections


def generate_vertical_bar_charts(
    base: alt.Chart,
    vertical_bar_charts_config: Dict[str, str],
    predictor: pd.DataFrame,
):
    vertical_bar_charts = {}
    x_variables_selections = {}
    for x_variable in vertical_bar_charts_config["x"]:
        if x_variable["field"] not in predictor.columns:
            continue
        x_variable_bar_charts = []
        x_variable_selection = alt.selection_point(fields=[x_variable["field"]])
        x_variables_selections[x_variable["field"]] = x_variable_selection
        x_variable_base_chart = (
            base.mark_bar()
            .encode(
                x=alt.X(**x_variable),
            )
            .add_params(x_variable_selection)
        )

        for y_variable in vertical_bar_charts_config["y"]:
            x_index_variable_color = alt.condition(
                x_variables_selections[x_variable["field"]],
                alt.Color(
                    "{}:N".format(x_variable["field"]),
                    legend=None,
                    sort=y_variable["sort"],
                ),
                alt.value("lightgray"),
            )
            x_variable_bar_chart = x_variable_base_chart.encode(
                y=y_variable["y"],
                color=x_index_variable_color,
                tooltip=y_variable["tooltip"],
            )

            x_variable_bar_charts.append(x_variable_bar_chart)
        vertical_bar_charts[x_variable["field"]] = x_variable_bar_charts
    return vertical_bar_charts, x_variables_selections


def add_interactive_selection(
    base: alt.Chart,
    selections: Dict[str, alt.SelectionParameter],
    selection_charts: Dict[str, List[alt.Chart]] = None,
):
    if selection_charts is None:
        selection_charts = {}
    for selection_variable, selection in selections.items():
        base = base.transform_filter(selection)
        for chart_variable in selection_charts.keys():
            if chart_variable != selection_variable:
                for i in range(len(selection_charts[chart_variable])):
                    selection_charts[chart_variable][i] = selection_charts[
                        chart_variable
                    ][i].transform_filter(selection)
    return base


def add_selection_on_legend(
    chart: alt.Chart, opacity_true: float = 1, opacity_false: float = 0.2
):
    legend_selection = alt.selection_point(fields=["value"], bind="legend")
    return chart.encode(
        opacity=alt.condition(
            legend_selection, alt.value(opacity_true), alt.value(opacity_false)
        )
    ).add_params(legend_selection)


def add_estimates_filters(
    base: alt.Chart,
    estimates_filters: Dict[str, alt.SelectionParameter],
    selection_charts: Dict[str, List[alt.Chart]] = None,
):
    if selection_charts is None:
        selection_charts = {}
    for estimate_filter in estimates_filters:
        base = base.transform_filter(estimate_filter)
        for chart_variable in selection_charts.keys():
            for i in range(len(selection_charts[chart_variable])):
                selection_charts[chart_variable][i] = selection_charts[chart_variable][
                    i
                ].transform_filter(estimate_filter)

    return base


def create_groupby_selection(
    indexes: List[Dict[str, str]],
    predictor: pd.DataFrame,
):
    index_fields = [
        index["field"] for index in indexes if index["field"] in predictor.columns
    ]
    if len(index_fields) >= 2:
        index_labels = [
            index["title"] for index in indexes if index["field"] in predictor.columns
        ]
        index_selection = alt.selection_point(
            fields=["index"],
            bind=alt.binding_radio(
                name="Group by: ", options=index_fields, labels=index_labels
            ),
            value=index_fields[0],
        )
    else:
        index_selection = None
    return index_selection, index_fields


def configure_style(
    chart: alt.Chart,
    chart_style: Dict[str, float],
):
    return chart.configure_axis(
        labelFontSize=chart_style["labelFontSize"],
        titleFontSize=chart_style["titleFontSize"],
        labelLimit=500,
    ).configure_legend(
        labelFontSize=chart_style["labelFontSize"],
        titleFontSize=chart_style["titleFontSize"],
        labelLimit=500,
    )


def concatenate_charts(
    main_chart: alt.Chart,
    horizontal_bar_charts: Dict[str, List[alt.Chart]],
    vertical_bar_charts: Dict[str, List[alt.Chart]],
    time_line: alt.Chart = None,
    spacing: float = 10,
):
    # Horizontal histograms
    bar_charts_rows = []
    for bar_charts_row in horizontal_bar_charts.values():
        bar_charts_row = reduce(
            lambda bar_chart_1, bar_chart_2: (bar_chart_1 | bar_chart_2).resolve_scale(
                color="independent"
            ),
            bar_charts_row,
        )
        bar_charts_rows.append(bar_charts_row)
    if bar_charts_rows:
        horizontal_bar_charts = reduce(
            lambda bar_chart_1, bar_chart_2: (bar_chart_1 & bar_chart_2).resolve_scale(
                color="independent"
            ),
            bar_charts_rows,
        )

    # Vertical histograms
    bar_charts_columns = []
    for bar_charts_column in vertical_bar_charts.values():
        bar_charts_column = reduce(
            lambda bar_chart_1, bar_chart_2: (bar_chart_1 | bar_chart_2).resolve_scale(
                color="independent"
            ),
            bar_charts_column,
        )
        bar_charts_columns.append(bar_charts_column)
    if bar_charts_columns:
        if bar_charts_rows:
            vertical_bar_charts = reduce(
                lambda bar_chart_1, bar_chart_2: (
                    bar_chart_1 & bar_chart_2
                ).resolve_scale(color="independent"),
                bar_charts_columns,
            )
        else:
            vertical_bar_charts = reduce(
                lambda bar_chart_1, bar_chart_2: (
                    bar_chart_1 | bar_chart_2
                ).resolve_scale(color="independent"),
                bar_charts_columns,
            )

    if bar_charts_rows:
        if bar_charts_columns:
            bottom_chart = vertical_bar_charts | horizontal_bar_charts
        else:
            bottom_chart = horizontal_bar_charts
    elif bar_charts_columns:
        bottom_chart = vertical_bar_charts
    else:
        bottom_chart = None

    if time_line:
        top_chart = alt.vconcat(main_chart, time_line, spacing=100).resolve_scale(
            color="independent"
        )
    else:
        top_chart = main_chart

    if bottom_chart:
        concat_chart = alt.vconcat(
            top_chart,
            bottom_chart,
            spacing=spacing,
        )
    else:
        concat_chart = top_chart
    return concat_chart


def month_diff(x, y):
    end = x.dt.to_period("M").view(dtype="int64")
    start = y.dt.to_period("M").view(dtype="int64")
    return end - start


def save_html(obj: alt.Chart, filename: str):
    """Save chart in the specified file

    Parameters
    ----------
    obj : alt.Chart
        Altair chart to be saved
    filename : str
        Folder path where to save the chart in HTML format.

        **EXAMPLE**: `"my_folder/my_file.html"`
    """
    if not isinstance(filename, Path):
        filename = Path(filename)
    Path.mkdir(filename.parent, exist_ok=True, parents=True)
    if hasattr(obj, "save"):
        obj.save(filename)
    else:
        with Path.open(filename, "w") as f:
            f.write(obj)
    logger.info("The chart has been saved in {}", filename)


def round_up(x: float, sig: int):
    if x == 0:
        return 0
    decimals = sig - floor(log10(abs(x))) - 1
    return ceil(x * 10**decimals) / 10**decimals


def scale_it(x: float):
    if x == 0:
        return 1
    return 10 ** ceil(log10(x))


def filter_predictor(
    predictor: pd.DataFrame,
    care_site_level: str = None,
    stay_type: List[str] = None,
    care_site_id: List[int] = None,
    care_site_short_name: List[int] = None,
    start_date: Union[datetime, str] = None,
    end_date: Union[datetime, str] = None,
    **kwargs
):
    # Time
    predictor = filter_table_by_date(
        table=predictor,
        table_name="predictor",
        start_date=start_date,
        end_date=end_date,
    )

    # Care site level
    if care_site_level:
        if not isinstance(care_site_level, list):
            care_site_level = [care_site_level]
        care_site_levels = []
        for care_site_lvl in care_site_level:
            if care_site_lvl in CARE_SITE_LEVEL_NAMES.keys():
                care_site_lvl = CARE_SITE_LEVEL_NAMES[care_site_lvl]
            if care_site_lvl not in predictor.care_site_level.unique():
                raise AttributeError(
                    "The selected care site level {} is not part of the computed care site levels {}".format(
                        care_site_level, list(predictor.care_site_level.unique())
                    )
                )
            care_site_levels.append(care_site_lvl)
        predictor = predictor[predictor.care_site_level.isin(care_site_levels)]
        logger.debug(
            "Predictor has been filtered on the selected care site level : {}",
            care_site_levels,
        )

    # Stay type
    if stay_type:
        if not isinstance(stay_type, list):
            stay_type = [stay_type]
        predictor = predictor[predictor.stay_type.isin(stay_type)]
        logger.debug(
            "Predictor has been filtered on the selected stay type : {}",
            stay_type,
        )

    # Care site id
    if care_site_id:
        if not isinstance(care_site_id, list):
            care_site_id = [care_site_id]
        predictor = predictor[predictor.care_site_id.isin(care_site_id)]
        logger.debug(
            "Predictor has been filtered on the selected care site id : {}",
            care_site_id,
        )

    # Care site short name
    if care_site_short_name:
        if not isinstance(care_site_short_name, list):
            care_site_short_name = [care_site_short_name]
        predictor = predictor[predictor.care_site_short_name.isin(care_site_short_name)]
        logger.debug(
            "Predictor has been filtered on the selected care site short name : {}",
            care_site_short_name,
        )

    # Others
    for key, value in kwargs.items():
        if not isinstance(value, list):
            value = [value]
        predictor = predictor[predictor[key].isin(value)]
        logger.debug(
            "Predictor has been filtered on the selected {} : {}",
            key,
            value,
        )

    # Care site specialty
    if (
        "care_site_specialty" in predictor.columns
        and predictor[~(predictor.care_site_specialty == "Non renseigné")].empty
    ):
        predictor = predictor.drop(columns="care_site_specialty")

    if predictor.empty:
        raise TypeError("Empty predictor: no data to plot.")
    return predictor



import catalogue

from edsteva.metrics.error import error
from edsteva.metrics.error_after_t0 import error_after_t0
from edsteva.metrics.error_between_t0_t1 import error_between_t0_t1

metrics = catalogue.create("edsteva", "metrics")

metrics.register("error", func=error)
metrics.register("error_after_t0", func=error_after_t0)
metrics.register("error_between_t0_t1", func=error_between_t0_t1)



from datetime import datetime, timedelta
from typing import Dict, List, Union

import numpy as np
import pandas as pd
from loguru import logger

from edsteva.utils.checks import check_columns
from edsteva.utils.framework import get_framework, to
from edsteva.utils.typing import DataFrame

from .utils import CARE_SITE_LEVEL_NAMES, get_child_and_parent_cs


def filter_table_by_type(
    table: DataFrame,
    table_name: str,
    type_groups: Union[str, Dict],
    source_col: str,
    target_col: str,
):
    if isinstance(type_groups, str):
        type_groups = {type_groups: type_groups}
    if isinstance(type_groups, dict):
        table_per_types = []
        for type_name, type_value in type_groups.items():
            table_per_type_element = table[
                table[source_col]
                .astype(str)
                .str.contains(
                    type_value,
                    case=False,
                    regex=True,
                    na=False,
                )
            ].copy()
            table_per_type_element[target_col] = type_name
            table_per_types.append(table_per_type_element)
    else:
        raise TypeError(
            "{} must be str or dict not {}".format(
                target_col, type(type_groups).__name__
            )
        )

    logger.debug(
        "The following {} : {} have been selected on table {}",
        target_col,
        type_groups,
        table_name,
    )
    return get_framework(table).concat(table_per_types, ignore_index=True)


def filter_valid_observations(
    table: DataFrame,
    table_name: str,
    invalid_naming: str = None,
    valid_naming: str = None,
):
    check_columns(
        df=table,
        required_columns=["row_status_source_value"],
    )

    if valid_naming:
        table_valid = table[table["row_status_source_value"] == valid_naming]
    elif invalid_naming:
        table_valid = table[~(table["row_status_source_value"] == invalid_naming)]
    else:
        raise Exception("valid_naming or invalid_naming must be provided.")
    table_valid = table_valid.drop(columns=["row_status_source_value"])
    logger.debug("Valid observations have been selected for table {}.", table_name)
    return table_valid


def filter_table_by_date(
    table: DataFrame,
    table_name: str,
    start_date: Union[datetime, str] = None,
    end_date: Union[datetime, str] = None,
):
    check_columns(df=table, required_columns=["date"])

    table = table.dropna(subset=["date"])
    logger.debug("Droping observations with missing date in table {}.", table_name)
    table["date"] = table["date"].astype("datetime64[ns]")
    start_date = pd.to_datetime(start_date)
    end_date = pd.to_datetime(end_date)

    if end_date and start_date:
        table = table[(table["date"] >= start_date) & (table["date"] < end_date)]
        logger.debug(
            "Observations between {} and {} have been selected for table {}.",
            start_date,
            end_date,
            table_name,
        )
    elif start_date:
        table = table[table["date"] >= start_date]
        logger.debug(
            "Observations after {} have been selected for table {}.",
            start_date,
            table_name,
        )
    elif end_date:
        table = table[table["date"] < end_date]
        logger.debug(
            "Observations before {} have been selected for table {}.",
            end_date,
            table_name,
        )
    # Truncate
    table["date"] = table["date"].dt.strftime("%Y-%m").astype("datetime64[ns]")
    return table


def filter_table_by_length_of_stay(
    visit_occurrence: DataFrame, length_of_stays: List[float]
):
    visit_occurrence = visit_occurrence.assign(
        length=(
            visit_occurrence.visit_end_datetime - visit_occurrence.visit_start_datetime
        )
        / np.timedelta64(timedelta(days=1))
    )

    # Incomplete stays
    visit_occurrence = visit_occurrence.assign(length_of_stay="Unknown")
    visit_occurrence["length_of_stay"] = visit_occurrence.length_of_stay.mask(
        visit_occurrence["visit_end_datetime"].isna(),
        "Incomplete stay",
    )

    # Complete stays
    min_duration = length_of_stays[0]
    max_duration = length_of_stays[-1]
    visit_occurrence["length_of_stay"] = visit_occurrence["length_of_stay"].mask(
        (visit_occurrence["length"] <= min_duration),
        "<= {} days".format(min_duration),
    )
    visit_occurrence["length_of_stay"] = visit_occurrence["length_of_stay"].mask(
        (visit_occurrence["length"] >= max_duration),
        ">= {} days".format(max_duration),
    )
    n_duration = len(length_of_stays)
    for i in range(0, n_duration - 1):
        min = length_of_stays[i]
        max = length_of_stays[i + 1]
        visit_occurrence["length_of_stay"] = visit_occurrence["length_of_stay"].mask(
            (visit_occurrence["length"] >= min) & (visit_occurrence["length"] < max),
            "{} days - {} days".format(min, max),
        )
    visit_occurrence = visit_occurrence.drop(columns="length")

    return visit_occurrence.drop(columns="visit_end_datetime")


def filter_table_by_care_site(
    table_to_filter: DataFrame,
    care_site_relationship: pd.DataFrame,
    care_site_ids: Union[int, List[int]] = None,
    care_site_short_names: Union[str, List[str]] = None,
    care_site_specialties: Union[str, List[str]] = None,
):
    care_site = care_site_relationship[
        [
            "care_site_id",
            "care_site_short_name",
            "care_site_level",
            "care_site_specialty",
        ]
    ]
    care_site_filter = []
    if care_site_ids:
        if not isinstance(care_site_ids, list):
            care_site_ids = [care_site_ids]
        care_site_filter.append(
            care_site[care_site["care_site_id"].isin(care_site_ids)]
        )
    if care_site_short_names:
        if not isinstance(care_site_short_names, list):
            care_site_short_names = [care_site_short_names]
        care_site_filter.append(
            care_site[care_site["care_site_short_name"].isin(care_site_short_names)]
        )

    if care_site_filter:
        care_site_filter = pd.concat(
            care_site_filter, ignore_index=True
        ).drop_duplicates()
        extended_care_site_id_to_filter = get_child_and_parent_cs(
            care_site_sample=care_site_filter,
            care_site_relationship=care_site_relationship,
        )
    else:
        extended_care_site_id_to_filter = care_site.copy()

    if care_site_specialties:
        if not isinstance(care_site_specialties, list):
            care_site_specialties = [care_site_specialties]
        extended_care_site_id_to_filter = extended_care_site_id_to_filter[
            extended_care_site_id_to_filter["care_site_specialty"].isin(
                care_site_specialties
            )
        ]
        extended_care_site_id_to_filter = get_child_and_parent_cs(
            care_site_sample=extended_care_site_id_to_filter,
            care_site_relationship=care_site_relationship,
        )

    extended_care_site_id_to_filter = list(
        extended_care_site_id_to_filter[
            (
                extended_care_site_id_to_filter.care_site_level.isin(
                    CARE_SITE_LEVEL_NAMES.values()
                )
            )
        ].care_site_id.unique()
    )

    return table_to_filter[
        table_to_filter.care_site_id.isin(extended_care_site_id_to_filter)
    ]


def convert_uf_to_pole(
    table: DataFrame,
    table_name: str,
    care_site_relationship: DataFrame,
):
    care_site_rel_uf_to_pole = _get_relationship_table_uf_to_hospital(
        care_site_relationship=care_site_relationship
    )[["care_site_id_uf", "care_site_id_pole"]].rename(
        columns={"care_site_id_uf": "care_site_id"}
    )
    care_site_rel_uf_to_pole = to(get_framework(table), care_site_rel_uf_to_pole)
    table = table.merge(care_site_rel_uf_to_pole, on="care_site_id", how="left")
    table["care_site_id"] = table["care_site_id_pole"].mask(
        table["care_site_id_pole"].isna(), table["care_site_id"]
    )
    table = table.drop(columns="care_site_id_pole")
    logger.debug(
        "For level {}, stays of the table {} located in {} have been linked to their corresponding {}",
        CARE_SITE_LEVEL_NAMES["Pole"],
        table_name,
        CARE_SITE_LEVEL_NAMES["UF"],
        CARE_SITE_LEVEL_NAMES["Pole"],
    )
    return table


def _get_relationship_table_uf_to_hospital(
    care_site_relationship: DataFrame,
):
    check_columns(
        df=care_site_relationship,
        required_columns=[
            "care_site_id",
            "care_site_level",
            "care_site_short_name",
            "parent_care_site_id",
            "parent_care_site_level",
            "parent_care_site_short_name",
        ],
    )

    care_site_rel_uf_pole = care_site_relationship.rename(
        columns={
            "care_site_level": "care_site_level_uf",
            "care_site_id": "care_site_id_uf",
            "care_site_short_name": "care_site_short_name_uf",
            "parent_care_site_level": "care_site_level_pole",
            "parent_care_site_id": "care_site_id_pole",
            "parent_care_site_short_name": "care_site_short_name_pole",
        }
    )

    uf_to_pole_care_site = care_site_rel_uf_pole[
        (care_site_rel_uf_pole["care_site_level_uf"] == CARE_SITE_LEVEL_NAMES["UF"])
    ]

    care_site_rel_pole_hospit = care_site_relationship.rename(
        columns={
            "care_site_id": "care_site_id_pole",
            "parent_care_site_level": "care_site_level_hospital",
            "parent_care_site_id": "care_site_id_hospital",
            "parent_care_site_short_name": "care_site_short_name_hospital",
        }
    ).drop(columns=["care_site_level", "care_site_short_name"])

    uf_to_hospital_care_site = uf_to_pole_care_site.merge(
        care_site_rel_pole_hospit, on="care_site_id_pole"
    )

    uf_to_hospital_care_site = uf_to_hospital_care_site[
        uf_to_hospital_care_site.care_site_level_pole == CARE_SITE_LEVEL_NAMES["Pole"]
    ]
    uf_to_hospital_care_site = uf_to_hospital_care_site[
        uf_to_hospital_care_site.care_site_level_hospital
        == CARE_SITE_LEVEL_NAMES["Hospital"]
    ]
    uf_to_hospital_care_site = uf_to_hospital_care_site.drop_duplicates(
        [
            "care_site_id_uf",
            "care_site_id_pole",
            "care_site_id_hospital",
        ]
    )
    return uf_to_hospital_care_site[
        [
            "care_site_id_uf",
            "care_site_short_name_uf",
            "care_site_id_pole",
            "care_site_short_name_pole",
            "care_site_id_hospital",
            "care_site_short_name_hospital",
        ]
    ]



from typing import Optional, Tuple

import pandas as pd
import pgpasslib
import psycopg2 as pg


class PostgresData:  # pragma: no cover
    """PostgreSQL interface to run SQL queries.

    This uses the file `~/.pgpass` to find the password and extra connection infos.

    Parameters
    ----------
    dbname : str
    schema : str
    user : str
    host : str
    port : str

    Examples
    --------
    >>> data = PostgresData(dbname="YOUR_DBNAME", schema="omop", user="YOUR_USERNAME")
    >>> data.read_sql("select count(*) from person")

    """

    def __init__(
        self,
        dbname: Optional[str] = None,
        schema: Optional[str] = None,
        user: Optional[str] = None,
        host: Optional[str] = None,
        port: Optional[int] = None,
    ):
        (
            self.host,
            self.port,
            self.dbname,
            self.user,
        ) = self._find_matching_pgpass_params(host, port, dbname, user)
        self.schema = schema

    @staticmethod
    def _find_matching_pgpass_params(
        host: str,
        port: int,
        dbname: str,
        user: str,
    ) -> Tuple:
        entries = pgpasslib._get_entries()
        consolidated_params = [
            (
                host or entry.host,
                port or entry.port,
                dbname or entry.dbname,
                user or entry.user,
            )
            for entry in entries
        ]
        matching_params = [
            params
            for entry, params in zip(entries, consolidated_params)
            if entry.match(*params)
        ]

        if len(matching_params) == 0:
            raise ValueError("Could not find matching entry in .pgpass file.")
        if len(matching_params) > 1:
            message = "\n".join(
                [
                    "Several entries found in .pgpass file. Be more specific.",
                    "The following entries match what you specified :",
                    *[str(params) for params in matching_params],
                ]
            )
            raise ValueError(message)

        return matching_params[0]

    def read_sql(self, sql_query: str, **kwargs) -> pd.DataFrame:
        """Execute pandas.read_sql() on the database.

        Parameters
        ----------
        sql_query : str
            SQL query (postgres flavor)
        **kwargs
            additional arguments passed to pandas.read_sql()

        Returns
        -------
        df : pandas.DataFrame

        """
        connection_infos = {
            param: getattr(self, param) for param in ["host", "port", "dbname", "user"]
        }
        connection_infos["password"] = pgpasslib.getpass(**connection_infos)
        connection = pg.connect(**connection_infos)
        if self.schema:
            connection.cursor().execute(f"SET SCHEMA '{self.schema}'")

        df = pd.read_sql(sql_query, con=connection, **kwargs)

        connection.close()
        return df



from typing import Callable, List

import numpy as np
import pandas as pd

from edsteva.utils.checks import check_columns
from edsteva.utils.loss_functions import l2_loss


def loss_minimization(
    predictor: pd.DataFrame,
    index: List[str],
    x_col: str = "date",
    y_col: str = "c",
    loss_function: Callable = l2_loss,
    min_rect_month_width=3,
):
    r"""Computes the threshold $t_0$ and $t_1$ of a predictor $c(t)$ by minimizing the following loss function:

    $$
    \begin{aligned}
    \mathcal{L}(t_0, t_1) & = \frac{\sum_{t = t_{min}}^{t_{max}} \mathcal{l}(c(t), f_{t_0, t_1}(t))}{t_{max} - t_{min}} \\
    (\hat{t_0}, \hat{t_1}) & = \underset{t_0, t_1}{\mathrm{argmin}}(\mathcal{L}(t_0, t_1))
    \end{aligned}
    $$

    Where the loss function $\mathcal{l}$ is by default the L2 distance and the estimated completeness $c_0$ is the mean completeness between $t_0$ and $t_1$.

    $$
    \begin{aligned}
    \mathcal{l}(c(t), f_{t_0, t_1}(t)) & = |c(t) - f_{t_0, t_1}(t)|^2 \\
    c_0 & = \frac{\sum_{t = t_0}^{t_1} c(t)}{t_1 - t_0}
    \end{aligned}
    $$



    Parameters
    ----------
    predictor : pd.DataFrame
        $c(t)$ computed in the Probe.
    index : List[str]
        Variable from which data is grouped.
        **EXAMPLE**: `["care_site_level", "stay_type", "note_type", "care_site_id"]`
    x_col : str, optional
        Column name for the time variable $t$.
    y_col : str, optional
        Column name  for the completeness variable $c(t)$.
    loss_function : Callable, optional
        The loss function $\mathcal{L}$.
    min_rect_month_width : int, optional
        Min number of months between $t_0$ and $t_1$.
    """
    check_columns(df=predictor, required_columns=[*index, x_col, y_col])
    predictor = predictor.sort_values(x_col)
    cols = [*index, x_col, y_col]
    iter = predictor[cols].groupby(index)
    results = []
    for partition, group in iter:
        row = dict(zip(index, partition))
        t_0, c_0, t_1 = _compute_one_double_threshold(
            group,
            x_col,
            y_col,
            loss_function,
            min_rect_month_width,
        )
        row["t_0"] = t_0
        row["c_0"] = c_0
        row["t_1"] = t_1
        results.append(row)

    return pd.DataFrame(results)


def _compute_one_double_threshold(
    group: pd.DataFrame,
    x_col: str,
    y_col: str,
    loss_func: Callable,
    min_rect_month_width: int,
):
    target = group[[x_col, y_col]].to_numpy()
    best_x0 = best_y0 = best_x1 = None
    best_loss = np.inf
    for idx in range(len(target) - min_rect_month_width):
        x0 = target[idx, 0]
        y_before_x0 = target[:idx, 1]
        for jdx in range(idx + min_rect_month_width, len(target)):
            x1 = target[jdx, 0]
            y_between_x0_x1 = target[idx:jdx, 1]
            y_after_x1 = target[jdx:, 1]
            y0 = y_between_x0_x1.mean()
            residual = np.hstack(
                [
                    y_before_x0,
                    (y_between_x0_x1 - y0),
                    y_after_x1,
                ]
            )
            loss = loss_func(residual).mean()
            if loss < best_loss:
                best_x0 = x0
                best_y0 = y0
                best_x1 = x1
                best_loss = loss

    return best_x0, best_y0, best_x1
